"""Find REAPER resource path without ``reapy_boost`` dist API enabled."""

import os
from typing import Generator

import reapy_boost
from .shared_library import is_windows, is_apple

if not reapy_boost.is_inside_reaper():
    # Third-party imports crash REAPER when run inside it.
    import psutil


def get_candidate_directories(
    detect_portable_install: bool = True
) -> Generator[str, None, None]:
    if detect_portable_install:
        yield get_portable_resource_directory()
    if is_apple():
        yield os.path.expanduser('~/Library/Application Support/REAPER')
    elif is_windows():
        yield os.path.expandvars(r'$APPDATA\REAPER')
    else:
        yield os.path.expanduser('~/.config/REAPER')


def get_portable_resource_directory() -> str:
    process_path = get_reaper_process_path()
    if is_apple():
        return '/'.join(process_path.split('/')[:-4])
    return os.path.dirname(process_path)


def get_reaper_process_path() -> str:
    """Return path to currently running REAPER process.

    Returns
    -------
    str
        Path to executable file.

    Raises
    ------
    RuntimeError
        When zero or more than one REAPER instances are currently
        running.
    """
    processes = [
        p for p in psutil.process_iter(['name', 'exe'])
        if os.path.splitext(p.info['name']  # type:ignore
                            )[0].lower() == 'reaper'
    ]
    if not processes:
        raise RuntimeError('No REAPER instance is currently running.')
    elif len(processes) > 1:
        raise RuntimeError(
            'More than one REAPER instance is currently running.'
        )
    return processes[0].info['exe']  # type:ignore


def get_resource_path(detect_portable_install: bool = True) -> str:
    for dir in get_candidate_directories(detect_portable_install):
        if os.path.exists(os.path.join(dir, 'reaper.ini')):
            return dir
    raise RuntimeError('Cannot find resource path')



from reapy_boost.errors import DisabledDistAPIError, UndefinedExtStateError
from reapy_boost.tools import json
from urllib import request
from urllib.error import URLError
from socket import timeout

import reapy_boost


class WebInterface:

    def __init__(self, port, host="localhost"):
        self._url = "http://{}:{}/_/".format(host, port)
        self.ext_state = ExtState(self)

    def activate_reapy_server(self):
        try:
            action_name = self.ext_state["activate_reapy_server"]
            self.perform_action(action_name)
        except UndefinedExtStateError:
            raise DisabledDistAPIError

    def get_reapy_server_port(self):
        try:
            port = self.ext_state["server_port"]
        except (URLError, timeout):
            raise DisabledDistAPIError
        except UndefinedExtStateError:
            self.activate_reapy_server()
            port = self.get_reapy_server_port()
        return port

    def perform_action(self, action_id):
        url = self._url + str(action_id)
        request.urlopen(url)


class ExtState:

    def __init__(self, web_interface):
        self._url = web_interface._url + "{method}/EXTSTATE/reapy_boost/{key}"

    def __getitem__(self, key):
        url = self._url.format(method="GET", key=key)
        string = request.urlopen(url, timeout=.5).read().decode("utf-8")
        value = string.split("\t")[-1][:-1]
        if not value:
            raise UndefinedExtStateError(key)
        value = json.loads(value)
        return value

    def __setitem__(self, key, value):
        value = json.dumps(value)
        url = self._url + "/{value}"
        url = url.format(method="SET", key=key, value=value)
        request.urlopen(url)



import contextlib
import functools
import typing as ty

import reapy_boost
import reapy_boost.config
from .network import machines
# if not reapy_boost.is_inside_reaper():
#     try:
#         from .network import Client, WebInterface
#         _WEB_INTERFACE = WebInterface(reapy_boost.config.WEB_INTERFACE_PORT)
#         _CLIENT = Client(_WEB_INTERFACE.get_reapy_server_port())
#     except DisabledDistAPIError:
#         import warnings
#         warnings.warn(DisabledDistAPIWarning())
#         _CLIENT = None

# from typing import Callable, Concatenate, TypeVar

FuncType = ty.Callable[..., ty.Any]
F = ty.TypeVar('F', bound=ty.Union[FuncType, property])


def dist_api_is_enabled() -> bool:
    """Return whether reapy_boost can reach REAPER from the outside."""
    return machines.get_selected_client() is not None


class inside_reaper(contextlib.ContextDecorator):
    """
    Context manager for efficient calls from outside REAPER.

    It can also be used as a function decorator.

    Examples
    --------
    Instead of running:

    >>> project = reapy_boost.Project()
    >>> l = [project.bpm for i in range(1000)

    which takes around 30 seconds, run:

    >>> project = reapy_boost.Project()
    >>> with reapy_boost.inside_reaper():
    ...     l = [project.bpm for i in range(1000)
    ...

    which takes 0.1 seconds!

    Example usage as decorator:

    >>> @reapy_boost.inside_reaper()
    ... def add_n_tracks(n):
    ...     for x in range(n):
    ...         reapy_boost.Project().add_track()

    """

    def __call__(
        self,
        func: F,
        encoded_func: ty.Optional[ty.Dict[str, object]] = None
    ) -> F:
        if reapy_boost.is_inside_reaper():
            return func
        if isinstance(func, property):
            return DistProperty.from_property(func)
        # Check if the decorated function is from reapy_boost
        module_name = func.__module__
        if module_name == 'reapy_boost' or module_name.startswith(
            'reapy_boost.'
        ):

            @functools.wraps(func)
            def wrap(*args, **kwargs):
                f = func if encoded_func is None else encoded_func
                client = machines.get_selected_client()
                return client.request(f, {"args": args, "kwargs": kwargs})

            return wrap
        # Otherwise, use the context manager
        return super().__call__(func)

    def __enter__(self) -> None:
        if not reapy_boost.is_inside_reaper():
            machines.get_selected_client().request("HOLD")

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        if not reapy_boost.is_inside_reaper():
            machines.get_selected_client().request("RELEASE")
        return False


class DistProperty(property):

    _inside_reaper = inside_reaper()

    @classmethod
    def from_property(cls, p: property) -> property:
        return cls().getter(p.fget).setter(p.fset).deleter(p.fdel)

    @staticmethod
    def _encode(f: F, method_name: str) -> ty.Dict[str, object]:
        return {
            "__callable__": True,
            "module_name": f.__module__,
            "name": "{}.f{}".format(f.__qualname__, method_name)
        }

    def getter(self, fget: F) -> property:
        if fget is not None:
            fget = self._inside_reaper(fget, self._encode(fget, "get"))
        return super().getter(fget)

    def setter(self, fset: F) -> property:
        if fset is not None:
            fset = self._inside_reaper(fset, self._encode(fset, "set"))
        return super().setter(fset)

    def deleter(self, fdel: F) -> property:
        if fdel is not None:
            fdel = self._inside_reaper(fdel, self._encode(fdel, "del"))
        return super().deleter(fdel)



import sys


def is_inside_reaper() -> bool:
    """
    Return whether ``reapy_boost`` is imported from inside REAPER.

    If ``reapy_boost`` is run from inside a REAPER instance but currently
    controls another REAPER instance on a slave machine (with
    ``reapy_boost.connect``), return False.
    """
    inside = hasattr(sys.modules["__main__"], "obj")
    if not inside:
        return False
    else:
        try:
            return machines.get_selected_machine_host() is None
        except NameError:
            # machines is undefined because we are still in the initial
            # import process.
            return True


from .tools import (
    connect,
    connect_to_default_machine,
    dist_api_is_enabled,
    inside_reaper,
    reconnect,
    generate_imgui,
)
from . import reascript_api, errors, ImGui
from .config import configure_reaper
from .core import *
from .core.reaper import *

__version__ = "0.10.101"
__all__ = [
    "reascript_api",
    "JS",
    "ImGui",
    "errors",
    # config
    "configure_reaper",
    # core.reapy_object
    "ReapyObject",
    "ReapyObjectList",
    # core.project
    "Marker",
    "Project",
    "Region",
    "TimeSelection",
    "MarkerInfo",
    "RegionInfo",
    # core.audio_accessor
    "AudioAccessor",
    # core.envelope
    "Envelope",
    "EnvelopeList",
    "EnvelopePoint",
    # core.fx
    "FX",
    "FXList",
    "FXParam",
    "FXParamsList",
    # core.item
    "Item",
    "MIDIEvent",
    "MIDIEventList",
    "CC",
    "CCList",
    "Note",
    "NoteList",
    "TextSysex",
    "TextSysexInfo",
    "TextSysexList",
    "CCShapeFlag",
    "CCShape",
    "MIDIEventDict",
    "MIDIEventInfo",
    "CCInfo",
    'NoteInfo',
    "Source",
    "Take",
    # core.track
    "AutomationItem",
    "Send",
    "Track",
    "TrackList",
    # core.window
    "MIDIEditor",
    "ToolTip",
    "Window",
    # core.reaper
    'add_reascript',
    'arm_command',
    'browse_for_file',
    'clear_console',
    'clear_peak_cache',
    'dB_to_slider',
    'delete_ext_state',
    'disarm_command',
    'get_armed_command',
    'get_command_id',
    'get_command_name',
    'get_exe_dir',
    'get_ext_state',
    'get_global_automation_mode',
    'get_ini_file',
    'get_last_touched_track',
    'get_main_window',
    'get_projects',
    'get_reaper_version',
    'get_resource_path',
    'get_user_inputs',
    'has_ext_state',
    'is_valid_id',
    'open_project',
    'perform_action',
    'prevent_ui_refresh',
    'print',
    'reaprint',
    'remove_reascript',
    'rgb_from_native',
    'rgb_to_native',
    'set_ext_state',
    'set_global_automation_mode',
    'show_console_message',
    'show_message_box',
    'slider_to_dB',
    'test_api',
    'undo_block',
    'update_arrange',
    'update_timeline',
    'view_prefs',
    'audio',
    'midi',
    'ui',
    'defer',
    'at_exit',
    # tools
    'connect',
    'connect_to_default_machine',
    'dist_api_is_enabled',
    'inside_reaper',
    'reconnect',
    'generate_imgui',
]



"""Define custom errors."""

from typing import Optional


class DisabledDistAPIError(Exception):

    def __init__(self) -> None:
        message = (
            "Can't reach distant API. Please start REAPER, or call " +
            "reapy_boost.config.enable_dist_api() from inside REAPER to enable "
            + "distant API."
        )
        super().__init__(message)


class DisabledDistAPIWarning(Warning):

    def __init__(self) -> None:
        message = (
            "Can't reach distant API. Please start REAPER, or call " +
            "reapy_boost.config.enable_dist_api() from inside REAPER to enable "
            + "distant API."
        )
        super().__init__(message)


class DisconnectedClientError(Exception):

    def __init__(self) -> None:
        message = "Client disconnected. Call self.connect to reconnect."
        super().__init__(message)


class DistError(Exception):

    def __init__(self, tb_string: str) -> None:
        message = (
            "\n\nAn error occurred while running a function inside REAPER. " +
            "Traceback was :\n\n{}"
        ).format(tb_string)
        super().__init__(message)


class ExtensionNotFoundError(Exception):

    def __init__(self, extension: str, url: str) -> None:
        message = (
            "Extension {} is required by this function " +
            "but is not available. " + "Please download it from {}."
        ).format(extension, url)
        super().__init__(message)


class InsideREAPERError(Exception):

    pass


class InvalidObjectError(Exception):
    """Raised when an object with invalid ID has tried to access REAPER.

    Common causes of this error are closing REAPER or deleting the
    object referred to by the aforementioned ID.

    The object that caused this error is available as its ``object``
    attribute.

    Parameters
    ----------
    object : ReapyObject
        Object that caused the error.

    Notes
    -----
    Most reapy_boost objects have a ``has_valid_id`` property that allows
    to check for its validity.
    """

    def __init__(self, object: object) -> None:
        self.object = object
        message = (
            "{} has an invalid ID. Common causes of this error " +
            "are closing REAPER or deleting the object referred to " +
            "by the aforementioned ID. Try checking for " +
            "object.has_valid_id`."
        )
        super().__init__(message.format(object))


class OutsideREAPERError(Exception):

    def __init__(self) -> None:
        message = "reapy_boost can not be enabled or disabled from outside REAPER"
        super().__init__(message)


class RedoError(Exception):

    def __init__(self) -> None:
        message = "Can't redo."
        super().__init__(message)


class UndefinedEnvelopeError(Exception):

    def __init__(
        self, index: Optional[int], name: Optional[str],
        chunk_name: Optional[str]
    ) -> None:
        if index is not None:
            message = "No envelope with index {}".format(index)
        elif name is not None:
            message = "No envelope with name {}".format(name)
        else:
            message = "No envelope with chunk name {}".format(chunk_name)
        super().__init__(message)


class UndefinedExtStateError(Exception):

    def __init__(self, key: str) -> None:
        message = "Undefined extended state for key {}.".format(key)
        super().__init__(message, key)


class UndefinedFXParamError(Exception):

    def __init__(self, fx_name: str, name: str) -> None:
        message = "No param named \"{}\" for FX \"{}\"".format(name, fx_name)
        super().__init__(message)


class UndefinedMarkerError(Exception):

    def __init__(self, index: int) -> None:
        message = "No marker with user-index {}".format(index)
        super().__init__(message)


class UndefinedRegionError(Exception):

    def __init__(self, index: int) -> None:
        message = "No region with user-index {}".format(index)
        super().__init__(message)


class UndoError(Exception):

    def __init__(self) -> None:
        message = "Can't undo."
        super().__init__(message)



import builtins
import functools
import typing as ty

import reapy_boost


@reapy_boost.inside_reaper()
def map(
    function: ty.Callable,
    *iterables: ty.Iterable,
    constants: ty.Mapping[str, ty.Any] = ...,
    kwargs_iterable: ty.Optional[ty.Iterable[ty.Mapping[str, ty.Any]]] = ...
) -> list:
    """Efficiently map a function to iterables of arguments.

    From outside REAPER, reapy_boost function calls are sent over the
    network for REAPER to execute them. The result in then sent back
    to the reapy_boost client.

    When calling a large number of reapy_boost functions, the network
    overhead can become significant. To solve this problem,
    ``reapy_boost.map`` sends all arguments at once and retrieves all
    results at once.

    It extends the API of the built-in Python function ``map`` with
    keyword arguments options.

    It stops when the shortest of ``kwargs_iterable`` and all
    iterables in ``iterables`` is exhausted.

    Parameters
    ----------
    function : reapy_boost callable
        Function to map to iterables of arguments.
    iterables : tuple of iterables
        Iterables of arguments that will be iterated in parallel and
        passed to ``function`` as positional arguments.
    constants : dict, optional
        Constant keyword arguments that will be passed  to
        ``function`` at each call.
    kwargs_iterable : iterable of mappings
        Iterable yielding mappings of keyword arguments that will be
        passed to ``function`` alongside the positional arguments in
        ``iterables``.

    Returns
    -------
    list
        The list of results of each calls. Unlike the built-in Python
        function ``map``, it does not return an iterator.

    Examples
    --------
    Below is a simple use case where ``reapy_boost.map`` reduces run time
    by 95%.

    >>> import time
    >>> import reapy_boost
    >>>
    >>> project = reapy_boost.Project()
    >>> take = project.items[0].active_take
    >>> with reapy_boost.inside_reaper():
    ...     start_time = time.time()
    ...     ppqs = [take.time_to_ppq(time) for time in range(10**5)]
    ...     print(f'Elapsed time without reapy_boost.map: {time.time() -\
                    start_time:.1f} s.')
    ...
    Elapsed time without reapy_boost.map: 15.0 s.
    >>>
    >>> start_time = time.time()
    >>> ppqs = reapy_boost.map(take.time_to_ppq, list(range(10**5)))
    >>> print(f'Elapsed time with reapy_boost.map: {time.time() - start_time:.1f}s.')
    Elapsed time with reapy_boost.map: 0.7s.
    """
    if kwargs_iterable is None:
        kwargs_iterable = iter(dict, None)
    partial_function = functools.partial(function, **constants)
    function = lambda *args: partial_function(*args[1:], **args[0])
    return list(builtins.map(function, kwargs_iterable, *iterables))



import struct
from typing import Iterable, List, Optional, Tuple, Union, cast
import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject


class Take(ReapyObject):

    _class_name = "Take"

    def __init__(self, id: str):
        self.id = id

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Take) and self.id == other.id

    @property
    def _args(self) -> Tuple[str]:
        return self.id,

    def add_audio_accessor(self) -> 'reapy_boost.AudioAccessor':
        """
        Create audio accessor and return it.

        Returns
        -------
        audio_accessor : AudioAccessor
            Audio accessor on take.
        """
        audio_accessor_id = RPR.CreateTakeAudioAccessor(self.id)  # type:ignore
        audio_accessor = reapy_boost.AudioAccessor(audio_accessor_id)
        return audio_accessor

    @reapy_boost.inside_reaper()
    def add_event(
        self,
        message: Iterable[int],
        position: float,
        unit: str = "seconds"
    ) -> None:
        """
        Add generic event to the take at position.

        Note
        ----
        ⋅ No sort events during this call
        ⋅ Inserting notes within this function causes problems
            (wrong note on and off timing), this is known REAPER bug.
            Use `Take.add_note` instead.

        Parameters
        ----------
        message : Iterable[int]
            Can be any message buffer, for example: (0xb0, 64, 127)
            which is CC64 val127 on channel 1
        position : float
            position at take
        unit : str, optional
            "beats"|"ppq"|"seconds" (default are seconds)

        See also
        --------
        Take.add_note
        """
        ppqpos = self._resolve_midi_unit((position, ), unit)[0]
        bytestr = self._midi_to_bytestr(message)
        RPR.MIDI_InsertEvt(  # type:ignore
            self.id, False, False, ppqpos, bytestr, len(bytestr)
        )

    def add_fx(
        self, name: str, even_if_exists: bool = True
    ) -> 'reapy_boost.FX':
        """
        Add FX to track and return it.

        Parameters
        ----------
        name : str
            FX name.
        even_if_exists : bool, optional
            Whether the FX should be added even if there already is an
            instance of the same FX on the track (default=True).

        Returns
        -------
        fx : FX
            New FX on take (or previously existing instance of FX if
            even_if_exists=False).

        Raises
        ------
        ValueError
            If there is no FX with the specified name.
        """
        index = RPR.TakeFX_AddByName(  # type:ignore
            self.id, name, 1 - 2 * even_if_exists)
        if index == -1:
            raise ValueError("Can't find FX named {}".format(name))
        fx = reapy_boost.FX(cast('reapy_boost.Take', self), index)
        return fx

    @reapy_boost.inside_reaper()
    def add_note(
        self,
        start: float,
        end: float,
        pitch: int,
        velocity: int = 100,
        channel: int = 0,
        selected: bool = False,
        muted: bool = False,
        unit: str = "seconds",
        sort: bool = True
    ) -> None:
        """
        Add MIDI note to take.

        Parameters
        ----------
        start : float
            Note start. Unit depends on ``unit``.
        end : float
            Note end. Unit depends on ``unit``.
        pitch : int
            Note pitch between 0 and 127.
        velocity : int, optional
            Note velocity between 0 and 127 (default=100).
        channel : int, optional
            MIDI channel between 0 and 15.
        selected : bool, optional
            Whether to select new note (default=False).
        muted : bool
            Whether to mute new note (default=False).
        unit : {"seconds", "ppq", "beats"}, optional
            Time unit for ``start`` and ``end`` (default="seconds").
            ``"ppq"`` refers to MIDI ticks.
        sort : bool, optional
            Whether to resort notes after creating new note
            (default=True). If False, then the new note will be
            ``take.notes[-1]``. Otherwise it will be at its place in
            the time-sorted list ``take.notes``. Set to False for
            improved efficiency when adding several notes, then call
            ``Take.sort_events`` at the end.

        See also
        --------
        Take.sort_events
        """
        start, end = self._resolve_midi_unit((start, end), unit)
        sort = bool(not sort)
        args = (
            self.id, selected, muted, start, end, channel, pitch, velocity,
            sort
        )
        RPR.MIDI_InsertNote(*args)  # type:ignore

    @reapy_boost.inside_reaper()
    def add_sysex(
        self,
        message: Iterable[int],
        position: float,
        unit: str = "seconds",
        evt_type: int = -1
    ) -> None:
        """
        Add SysEx event to take.

        Notes
        -----
        ⋅ No sort events during this call
        ⋅ No need to add 0xf0 ... 0xf7 bytes (they will be doubled)

        Parameters
        ----------
        message : Iterable[int]
            Can be any message buffer, for example: (0xb0, 64, 127)
            which is CC64 val127 on channel 1
        position : float
            position at take
        unit : str, optional
            "beats"|"ppq"|"seconds" (default are seconds)
        evt_type: int (default -1)
            Allowable types are
            ⋅ -1:sysex (msg should not include bounding F0..F7),
            ⋅ 1-14:MIDI text event types,
            ⋅ 15=REAPER notation event.
        """
        bytestr = self._midi_to_bytestr(message)
        ppqpos = self._resolve_midi_unit((position, ), unit)[0]
        RPR.MIDI_InsertTextSysexEvt(  # type:ignore
            self.id, False, False, ppqpos, evt_type, bytestr, len(bytestr)
        )

    def beat_to_ppq(self, beat: float) -> int:
        """
        Convert beat number (from project start) to MIDI ticks (of the take).

        Parameters
        ----------
        beat : float
            Beat time to convert in beats.

        Returns
        -------
        ppq : float
            Converted time in MIDI ticks of current take.

        See also
        --------
        Take.ppq_to_beat
        Take.time_to_ppq
        """
        ppq = RPR.MIDI_GetPPQPosFromProjQN(self.id, beat)  # type:ignore
        return int(ppq)

    @property
    def cc_events(self) -> 'reapy_boost.CCList':
        """
        List of CC events on take.

        :type: CCList
        """
        return reapy_boost.CCList(cast('reapy_boost.Take', self))

    @property
    def envelopes(self) -> 'reapy_boost.EnvelopeList':
        return reapy_boost.EnvelopeList(cast('reapy_boost.Take', self))

    @property
    def fxs(self) -> 'reapy_boost.FXList':
        """
        FXs on take.

        :type: FXList
        """
        return reapy_boost.FXList(cast('reapy_boost.Take', self))

    def get_info_value(self, param_name):
        return RPR.GetMediaItemTakeInfo_Value(  # type:ignore
            self.id, param_name)

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if take has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        return self._has_valid_id_inside()

    @reapy_boost.inside_reaper()
    def _has_valid_id_inside(self) -> bool:
        try:
            project_id = self.track.project.id
        except (OSError, AttributeError):
            return False
        pointer, name = self._get_pointer_and_name()
        return bool(RPR.ValidatePtr2(project_id, pointer, name))

    @property
    def is_active(self) -> bool:
        """
        Whether take is active.

        :type: bool
        """
        return self._is_active_inside()

    @reapy_boost.inside_reaper()
    def _is_active_inside(self) -> bool:
        return self == self.item.active_take

    @property
    def is_midi(self) -> bool:
        """
        Whether take contains MIDI or audio.

        :type: bool
        """
        return bool(RPR.TakeIsMIDI(self.id))  # type:ignore

    @property
    def item(self) -> 'reapy_boost.Item':
        """
        Parent item.

        :type: Item
        """
        return reapy_boost.Item(
            cast(
                str,
                RPR.GetMediaItemTake_Item(  # type:ignore
                    self.id
                )
            )
        )

    @reapy_boost.inside_reaper()
    def get_midi(
        self,
        size: int = 2 * 1024 * 1024
    ) -> List['reapy_boost.MIDIEventDict']:
        """
        Get all midi data of take in one call.

        Parameters
        ----------
        size : int, optional
            predicted size of event buffer to allocate.
            For performance is better to set it to something a but bigger
            than expected buffer size.
            Event with common midi-message (3-bytes) takes 12 bytes.

        Returns
        -------
        List[MIDIEventDict]
            ppq: int
            selected: bool
            muted: bool
            cc_shape: CCShapeFlag
            buf: ty.List[int]

        See also
        --------
        Take.set_midi
        """
        buf: str
        _, _, buf, _ = RPR.MIDI_GetAllEvts(  # type:ignore
            self.id, '', size
        )
        msg = buf.encode('latin-1')
        out = []
        i = 0
        ppq = 0
        length = len(msg)
        while i < length:
            if length - i < 9:
                break
            ofst, flag, len_ = (
                struct.unpack('<I', msg[i:i + 4])[0], int(msg[i + 4]),
                struct.unpack('<I', msg[i + 5:i + 9])[0]
            )
            ppq += ofst
            buf_b = msg[i + 9:i + 9 + len_]
            buf_int = [int(b) for b in buf_b]
            i += 9 + len_
            if len_ == 0:
                i += 1
                continue
            out.append(
                reapy_boost.MIDIEventDict(
                    ppq=ppq,
                    selected=bool(flag & 1),
                    muted=bool(flag & 2),
                    cc_shape=reapy_boost.CCShapeFlag(flag & 0b11110000),
                    buf=buf_int
                )
            )
        return out

    @property
    def guid(self) -> str:
        """
        Used for communication within other scripts.

        :type: str
        """
        _, _, _, guid, _ = RPR.GetSetMediaItemTakeInfo_String(  # type:ignore
            self.id, 'GUID', 'stringNeedBig', False
        )
        return guid

    def make_active_take(self) -> bool:
        """
        Make take active.
        """
        RPR.SetActiveTake(self.id)  # type:ignore

    @property
    def midi_events(self) -> 'reapy_boost.MIDIEventList':
        """
        Get all midi events as EventList.

        Returns
        -------
        MIDIEventList
        """
        return reapy_boost.MIDIEventList(cast('reapy_boost.Take', self))

    def midi_hash(self, notes_only: bool = False) -> str:
        """
        Get hash of MIDI-data to compare with later.

        Parameters
        ----------
        notes_only : bool, (False by default)
            count only notes if True

        Returns
        -------
        str
        """
        return RPR.MIDI_GetHash(  # type:ignore
            self.id, notes_only, 'hash', 1024**2)[3]

    def _midi_to_bytestr(self, message: Iterable[int]) -> str:
        return bytes(message).decode('latin-1')

    @property
    def n_cc(self) -> int:
        """
        Number of MIDI CC events in take (always 0 for audio takes).

        :type: int
        """
        return RPR.MIDI_CountEvts(self.id, 0, 0, 0)[3]  # type:ignore

    @property
    def n_envelopes(self) -> int:
        """
        Number of envelopes on take.

        :type: int
        """
        return RPR.CountTakeEnvelopes(self.id)  # type:ignore

    @property
    def n_fxs(self) -> int:
        """
        Number of FXs on take.

        :type: int
        """
        return RPR.TakeFX_GetCount(self.id)  # type:ignore

    @property
    def n_midi_events(self) -> int:
        """
        Number of MIDI events in take.

        :type: int
        """
        return RPR.MIDI_CountEvts(self.id, 1, 1, 1)[0]  # type:ignore

    @property
    def n_notes(self) -> int:
        """
        Number of MIDI notes in take (always 0 for audio takes).

        :type: int
        """
        return RPR.MIDI_CountEvts(self.id, 0, 0, 0)[2]  # type:ignore

    @property
    def n_text_sysex(self) -> int:
        """
        Number of MIDI text/sysex events in take (0 for audio takes).

        :type: int
        """
        return RPR.MIDI_CountEvts(self.id, 0, 0, 0)[4]  # type:ignore

    @property
    def name(self) -> str:
        """
        Take name.

        :type: str
        """
        if self._is_defined:
            return RPR.GetTakeName(self.id)  # type:ignore
        return ""

    @name.setter
    def name(self, name: str) -> None:
        RPR.GetSetMediaItemTakeInfo_String(  # type:ignore
            self.id, "P_NAME", name, True)

    @property
    def notes(self) -> 'reapy_boost.NoteList':
        """
        List of MIDI notes on take.

        Unless ``Take.add_note`` has been called with ``sort=False``,
        notes are time-sorted.

        :type: NoteList
        """
        return reapy_boost.NoteList(cast('reapy_boost.Take', self))

    def ppq_to_beat(self, ppq: Union[int, float]) -> float:
        """
        Convert time in MIDI ticks to beats.

        Note
        ----
        Ticks are counted from take start, beats — from project start.

        Parameters
        ----------
        ppq : float
            Time to convert in MIDI ticks.

        Returns
        -------
        beat : float
            Converted time in beats.

        See also
        --------
        Take.beat_to_ppq
        Take.ppq_to_time
        """
        beat = RPR.MIDI_GetProjQNFromPPQPos(self.id, ppq)  # type:ignore
        return beat

    def ppq_to_time(self, ppq: Union[int, float]) -> float:
        """
        Convert time in MIDI ticks to seconds.

        Parameters
        ----------
        ppq : float
            Time to convert in MIDI ticks.

        Returns
        -------
        time : float
            Converted time in seconds.

        See also
        --------
        Take.time_to_ppq
        """
        time = RPR.MIDI_GetProjTimeFromPPQPos(self.id, ppq)  # type:ignore
        return time

    @property
    def project(self) -> 'reapy_boost.Project':
        """
        Take parent project.

        :type: reapy_boost.Project
        """
        return self._project_inside()

    @reapy_boost.inside_reaper()
    def _project_inside(self) -> 'reapy_boost.Project':
        return self.item.project

    @reapy_boost.inside_reaper()
    def _resolve_midi_unit(
        self,
        pos_tuple: Tuple[float, ...],
        unit: str = "seconds"
    ) -> Tuple[float, ...]:
        """Get positions as ppq from tuple of positions of any length.

        Parameters
        ----------
        pos_tuple : Tuple[float, ...]
            tuple of position time in bets, ppq or seconds.
        unit : str, optional
            type of position inside tuple: seconds|beats|ppq

        Returns
        -------
        Tuple[float, ...]
            the same tuple normalized to ppq
        """
        if unit == "ppq":
            return pos_tuple
        # item_start_seconds = self.item.position

        def resolver(pos):
            if unit == "beats":
                return self.beat_to_ppq(pos)
            if unit == "seconds":
                return self.time_to_ppq(pos)
            raise ValueError('unit param should be one of seconds|beats|ppq')

        return cast(Tuple[int, int], (resolver(pos) for pos in pos_tuple))

    def select_all_midi_events(self, select: bool = True) -> None:
        """
        Select or unselect all MIDI events.

        Parameters
        ----------
        select : bool
            Whether to select or unselect events.

        See also
        --------
        Take.unselect_all_midi_events
        """
        RPR.MIDI_SelectAll(self.id, select)  # type:ignore

    def set_info_value(self, param_name, value):
        return RPR.SetMediaItemTakeInfo_Value(  # type:ignore
            self.id, param_name, value)

    @reapy_boost.inside_reaper()
    def set_midi(
        self,
        midi: List['reapy_boost.MIDIEventDict'],
        start: Optional[float] = None,
        unit: str = "seconds",
        sort: bool = True
    ):
        """
        Erase all midi from take and build new one from scratch.

        Parameters
        ----------
        midi : List[MIDIEventDict]
            can be taken with `Take.get_midi()` or build from scratch.
        start : float, optional
            if offset needed (for example, start from a particular time)
        unit : str, optional
            time unit: "seconds"|"beats"|"ppq"
        sort : bool, optional
            if sort is needed after insertion

        Raises
        ------
        NotImplementedError
            currently, gaps between events longer than several hours
            are not supported.

        See also
        --------
        Take.get_midi
        """
        out = b''
        start_ppq = 0
        if start is not None:
            start_ppq = self._resolve_midi_unit((start, ), unit)[0]
        last_ppq = 0 - start_ppq
        for msg in midi:
            evt = b''
            ofst_i = msg['ppq'] - last_ppq
            if ofst_i > 4294967295:
                raise NotImplementedError(
                    'ofset larger than 4294967295 currently not supported'
                )
                # something done with big offset
                # it is about many-many-many hours between events
                # (1 hour is about 8000000 ppq in 120 bpm)
            ofst = struct.pack('<I', int(ofst_i))
            evt += ofst
            last_ppq = msg['ppq']

            flag = bytes(
                [
                    int(msg['selected']) | (int(msg['muted']) << 1)
                    | msg['cc_shape']
                ]
            )
            evt += flag

            len_ = struct.pack('<I', len(msg['buf']))
            evt += len_

            buf = bytes(msg['buf'])
            evt += buf
            out += evt
        packed = out.decode('latin-1')
        RPR.MIDI_SetAllEvts(self.id, packed, len(packed))  # type:ignore
        if sort:
            self.sort_events()

    def sort_events(self) -> None:
        """
        Sort MIDI events on take.

        This is only needed if ``Take.add_note`` was called with
        ``sort=False``.

        Examples
        --------
        The following example creates 100 MIDI notes on take in
        reversed order, with ``sort=False`` for efficiency. Thus,
        ``take.notes`` is not time-sorted. ``take.sort_events`` is
        called afterwards so that ``take.notes`` is time-sorted.

        >>> for i in range(100):
        ...     take.add_note(99 - i, 100 - i, pitch=0, sort=False)
        ...
        >>> take.notes[0].start, take.notes[1].start
        99.0, 98.0
        >>> take.sort_events()
        >>> take.notes[0].start, take.notes[1].start
        0.0, 1.0
        """
        RPR.MIDI_Sort(self.id)  # type:ignore

    @property
    def source(self) -> 'reapy_boost.Source':
        """
        Take source.

        :type: Source
        """
        return reapy_boost.Source(
            cast(
                str,
                RPR.GetMediaItemTake_Source(  # type:ignore
                    self.id
                )
            )
        )

    @source.setter
    def source(self, source: 'reapy_boost.Source') -> None:
        RPR.SetMediaItemTake_Source(self.id, source.id)  # type:ignore

    @property
    def start_offset(self) -> float:
        """
        Start time of the take relative to start of source file.

        :type: float
        """
        return self.get_info_value("D_STARTOFFS")

    @start_offset.setter
    def start_offset(self, value: float) -> None:
        self.set_info_value("D_STARTOFFS", value)

    @property
    def text_sysex_events(self) -> 'reapy_boost.TextSysexList':
        """
        List of text or SysEx events.

        :type: TextSysexList
        """
        return reapy_boost.TextSysexList(cast('reapy_boost.Take', self))

    def time_to_ppq(self, time: float) -> int:
        """
        Convert time in seconds to MIDI ticks.

        Parameters
        ----------
        time : float
            Time to convert in seconds.

        Returns
        -------
        ppq : float
            Converted time in MIDI ticks.

        See also
        --------
        Take.ppq_to_time
        """
        ppq = RPR.MIDI_GetPPQPosFromProjTime(self.id, time)  # type:ignore
        return int(ppq)

    @property
    def track(self) -> 'reapy_boost.Track':
        """
        Parent track of take.

        :type: Track
        """
        track_id = RPR.GetMediaItemTake_Track(self.id)  # type:ignore
        return reapy_boost.Track(cast(str, track_id))

    def unselect_all_midi_events(self) -> None:
        """
        Unselect all MIDI events.

        See also
        --------
        Take.select_all_midi_events
        """
        self.select_all_midi_events(select=False)

    @property
    def visible_fx(self) -> Optional['reapy_boost.FX']:
        """
        Visible FX in FX chain if any, else None.

        :type: FX or NoneType
        """
        with reapy_boost.inside_reaper():
            return self.fxs[cast(
                int,
                RPR.TakeFX_GetChainVisible(  # type:ignore
                    self.id
                )
            )]



from .config import *
from .shared_library import get_python_shared_library

__all__ = [
    'add_web_interface',
    'configure_reaper',
    'create_new_web_interface',
    'delete_web_interface',
    'disable_dist_api',
    'enable_dist_api',
    'enable_python',
    'REAPY_SERVER_PORT',
    'WEB_INTERFACE_PORT',
    "get_python_shared_library",
]



import reapy_boost
import reapy_boost.reascript_api as RPR
import typing as ty


def get_active_editor() -> ty.Optional[reapy_boost.MIDIEditor]:
    """
    Return active MIDI editor, or None if no editor is active.

    Returns
    -------
    editor : MIDIEditor or None
        Active MIDI editor, or None if no editor is active.
    """
    editor = reapy_boost.MIDIEditor(RPR.MIDIEditor_GetActive())  # type:ignore
    if not editor._is_defined:
        editor = None
    return editor


@reapy_boost.inside_reaper()
def get_input_names() -> ty.List[str]:
    """
    Return names of all input channels.

    Returns
    -------
    names : list of str
        Names of input channels.
    """
    n_channels = reapy_boost.midi.get_n_inputs()
    return [
        RPR.GetMIDIInputName(  # type:ignore
            i, "", 2048
        )[2] for i in range(n_channels)
    ]


def get_max_inputs() -> int:
    """
    Return maximum number of MIDI inputs.

    Returns
    -------
    max_inputs : int
        Maximum number of MIDI inputs.
    """
    max_inputs = RPR.GetMaxMidiInputs()  # type:ignore
    return max_inputs


def get_max_outputs() -> int:
    """
    Return maximum number of MIDI outputs.

    Returns
    -------
    max_outputs : int
        Maximum number of MIDI outputs.
    """
    max_outputs = RPR.GetMaxMidiOutputs()  # type:ignore
    return max_outputs


def get_n_inputs() -> int:
    """
    Return number of MIDI inputs.

    Returns
    -------
    n_inputs : int
        Number of MIDI inputs.
    """
    n_inputs = RPR.GetNumMIDIInputs()  # type:ignore
    return n_inputs


def get_n_outputs() -> int:
    """
    Return number of MIDI outputs.

    Returns
    -------
    n_outputs : int
        Number of MIDI outputs.
    """
    n_outputs = RPR.GetNumMIDIOutputs()  # type:ignore
    return n_outputs


@reapy_boost.inside_reaper()
def get_output_names() -> ty.List[str]:
    """
    Return names of all output channels.

    Returns
    -------
    names : list of str
        Names of output channels.
    """
    n_channels = reapy_boost.midi.get_n_outputs()
    return [
        RPR.GetMIDIOutputName(  # type:ignore
            i, "", 2048
        )[2] for i in range(n_channels)
    ]


def reinit() -> None:
    """Reset all MIDI devices."""
    RPR.midi_reinit()  # type:ignore



"""Define Server class."""

import reapy_boost
from reapy_boost.tools import json
from .socket import Socket

import socket
import traceback


class Server(Socket):
    """
    Server part of the ``reapy_boost`` dist API.

    It is instantiated inside REAPER. It receives and processes API
    call requests coming from the outside.
    """

    def __init__(self, port):
        super().__init__()
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.bind(("0.0.0.0", port))
        self.listen()
        self.connections = {}
        self.settimeout(.0001)

    @Socket._non_blocking
    def _get_request(self, connection, address):
        try:
            request = connection.recv()
            request = json.loads(request.decode())
        except (ConnectionAbortedError, ConnectionResetError):
            # Client has disconnected
            # Pretend client has nicely requested to disconnect
            input = {"args": (address, ), "kwargs": {}}
            request = {"function": self.disconnect, "input": input}
        return request

    def _hold_connection(self, address):
        connection = self.connections[address]
        result = {"type": "result", "value": None}
        self._send_result(connection, result)
        request = self._get_request(connection, address)
        while request is None or request["function"] != "RELEASE":
            if request is None:
                request = self._get_request(connection, address)
                continue
            result = self._process_request(request, address)
            try:
                self._send_result(connection, result)
                request = self._get_request(connection, address)
            except (ConnectionAbortedError, ConnectionResetError):
                # request was to disconnect
                request = {"function": "RELEASE"}
        result = {"type": "result", "value": None}
        return result

    def _process_request(self, request, address):
        if request["function"] == "HOLD":
            return self._hold_connection(address)
        args, kwargs = request["input"]["args"], request["input"]["kwargs"]
        result = {}
        try:
            result["value"] = request["function"](*args, **kwargs)
            result["type"] = "result"
        except Exception:
            # Errors are sent back to the client instead of raised in REAPER
            # (which would cause the server to crash).
            result["traceback"] = traceback.format_exc()
            result["type"] = "error"
        return result

    def _send_result(self, connection, result):
        result = json.dumps(result).encode()
        connection.send(result)

    @Socket._non_blocking
    def accept(self, *args, **kwargs):
        connection, address = super().accept()
        self.connections[address] = connection
        connection.send("{}".format(address).encode("ascii"))
        return connection, address

    def disconnect(self, address):
        connection = self.connections[address]
        connection.shutdown(socket.SHUT_RDWR)
        connection.close()
        del self.connections[address]

    def get_requests(self):
        requests = {}
        for address, connection in self.connections.items():
            request = self._get_request(connection, address)
            if request is not None:
                requests[address] = request
        return requests

    def process_requests(self, requests):
        results = {}
        for address, request in requests.items():
            result = self._process_request(request, address)
            results[address] = result
        return results

    def send_results(self, results):
        for address, result in results.items():
            try:
                connection = self.connections[address]
                self._send_result(connection, result)
            except (
                KeyError, BrokenPipeError, ConnectionAbortedError,
                ConnectionResetError
            ):
                # Happens when the client requested to disconnect.
                # Nothing must be returned in that case.
                pass



from typing import Any, Dict, Tuple, Type, TypedDict
import reapy_boost
import reapy_boost.reascript_api as RPR

ReapyObjectDict = TypedDict(
    'ReapyObjectDict', {
        "__reapy__": bool,
        "class": str,
        "args": Tuple[Any, ...],
        "kwargs": Dict[str, Any]
    }
)


class ReapyMetaclass(type):

    @property
    def _reapy_parent(self) -> Type['ReapyObject']:
        """Return first reapy_boost parent class."""
        for candidate in self.__mro__:
            if candidate.__module__.startswith('reapy_boost.'):
                return candidate
        raise TypeError(f'cannot find reapy_boost class in {self.__mro__}')


class ReapyObject(metaclass=ReapyMetaclass):
    """Base class for reapy_boost objects."""
    id: str

    def __eq__(self, other: object) -> bool:
        return repr(self) == repr(other)

    def __repr__(self) -> str:

        def to_str(x):
            if isinstance(x, str):
                return "\"{}\"".format(x)
            return str(x)

        args = ", ".join(map(to_str, self._args))
        kwargs = ", ".join(
            ("{}={}".format(k, to_str(v)) for k, v in self._kwargs.items())
        )
        if args and kwargs:
            brackets = ", ".join((args, kwargs))
        else:
            brackets = args if args else kwargs
        rep = "{}({})".format(self.__class__.__name__, brackets)
        return rep

    @property
    def _args(self) -> Tuple[()]:
        return ()

    def _get_pointer_and_name(self) -> Tuple[int, str]:
        name, pointer = self.id.split(')')
        return int(pointer, base=16), name[1:]

    @property
    def _is_defined(self) -> bool:
        if hasattr(self, "id"):
            return not self.id.endswith("0x0000000000000000")
        raise NotImplementedError

    @property
    def _kwargs(self) -> Dict[str, object]:
        return {}

    def _to_dict(self) -> ReapyObjectDict:
        return {
            "__reapy__": True,
            "class": self.__class__._reapy_parent.__name__,
            "args": self._args,
            "kwargs": self._kwargs
        }


class ReapyObjectList(ReapyObject):
    """Abstract class for list of ReapyObjects."""

    pass



from .item import Item
from .midi_event import (
    MIDIEvent, MIDIEventList, CC, CCList, Note, NoteList, TextSysex,
    TextSysexInfo, TextSysexList, CCShapeFlag, CCShape, MIDIEventDict,
    MIDIEventInfo, CCInfo, NoteInfo
)
from .source import Source
from .take import Take

__all__ = [
    'Item',
    'MIDIEvent',
    'MIDIEventList',
    'CC',
    'CCList',
    'Note',
    'NoteList',
    'TextSysex',
    'TextSysexInfo',
    'TextSysexList',
    'CCShapeFlag',
    'CCShape',
    'MIDIEventDict',
    'MIDIEventInfo',
    'CCInfo',
    'NoteInfo',
    'Source',
    'Take',
]


from typing import Tuple, cast
import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject


class Source(ReapyObject):

    def __init__(self, id: str) -> None:
        self.id = id

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Source) and self.id == other.id

    @property
    def _args(self) -> Tuple[str]:
        return self.id,

    def delete(self) -> None:
        """
        Delete source. Be sure that no references to source remains.
        """
        RPR.PCM_Source_Destroy(self.id)  # type:ignore

    @property
    def filename(self) -> str:
        """
        Return source file name.

        Returns
        -------
        filename : str
            Source file name.
        """
        _, filename, _ = RPR.GetMediaSourceFileName(  # type:ignore
            self.id, "", 10**5)
        return filename

    @property
    def has_valid_id(self) -> bool:
        """
        Whether ReaScript ID is still valid.

        For instance, if source has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        return self._has_valid_id_inside()

    @reapy_boost.inside_reaper()
    def _has_valid_id_inside(self) -> bool:
        pointer, name = self._get_pointer_and_name()
        return any(
            RPR.ValidatePtr2(project.id, pointer, name)  # type:ignore
            for project in reapy_boost.get_projects()
        )

    def length(self, unit: str = "seconds") -> float:
        """
        Return source length in `unit`.

        Parameters
        ----------
        unit : {"beats", "seconds"}

        Returns
        -------
        length : float
            Source length in `unit`.
        """
        length, _, is_quantized = RPR.GetMediaSourceLength(  # type:ignore
            self.id, 0)
        length = cast(float, length)
        if is_quantized:
            if unit == "beats":
                return length
            else:
                # elif unit == "seconds":
                raise NotImplementedError()
        else:
            if unit == "beats":
                raise NotImplementedError()
            else:
                # elif unit == "seconds":
                return length

    @property
    def n_channels(self) -> int:
        """
        Return number of channels in source media.

        Returns
        -------
        n_channels : int
            Number of channels in source media.
        """
        n_channels = RPR.GetMediaSourceNumChannels(self.id)  # type:ignore
        return n_channels

    @property
    def sample_rate(self) -> int:
        """
        Return source sample rate.

        Returns
        -------
        sample_rate : int
            Source sample rate.
        """
        sample_rate = RPR.GetMediaSourceSampleRate(self.id)  # type:ignore
        return sample_rate

    @property
    def type(self) -> str:
        """
        Return source type ("WAV, "MIDI", etc.).

        Returns
        -------
        type : str
            Source type.
        """
        _, type, _ = RPR.GetMediaSourceType(self.id, "", 10**5)  # type:ignore
        return type



import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject, ReapyObjectList
from reapy_boost.core.project.project import Project
from reapy_boost.core.audio_accessor import AudioAccessor
from reapy_boost.core.fx.fx import FX
from reapy_boost.core.item.item import Item
from reapy_boost.errors import InsideREAPERError, InvalidObjectError, UndefinedEnvelopeError
import typing as ty


class Track(ReapyObject):
    """
    REAPER Track.

    Parameters
    ----------
    id : str or int
        If str, can either be a ReaScript ID (usually looking like
        ``"(MediaTrack*)0x00000000110A1AD0"``), or a track name. In
        that case, ``project`` must be specified.
        If int, the index of the track. In that case, ``project`` must
        be specified.
    project : Project
        Parent project of the track. Only necessary to retrieve a
        track from its name or index.

    Examples
    --------
    In most cases, accessing tracks is better done directly from
    the parent Project:

    >>> project = reapy_boost.Project()
    >>> project.tracks[0]
    Track("(MediaTrack*)0x00000000110A1AD0")
    >>> project.tracks["PIANO"]  # This is actually the same track
    Track("(MediaTrack*)0x00000000110A1AD0")

    But the same track can also directly be instantiated with:

    >>> reapy_boost.Track(0, project)
    Track("(MediaTrack*)0x00000000110A1AD0")

    or

    >>> reapy_boost.Track("PIANO", project)
    Track("(MediaTrack*)0x00000000110A1AD0")
    """

    def __init__(
        self,
        id: ty.Union[str, int],
        project: ty.Optional[Project] = None
    ) -> None:
        self._project = None
        if isinstance(id, int):  # id is a track index
            id = ty.cast(str, RPR.GetTrack(project.id, id))  # type:ignore
            if id.endswith("0x0000000000000000"):
                raise IndexError('Track index out of range')
            self._project = project
        elif isinstance(id, str) and not id.startswith("(MediaTrack*)"):
            # id is a track name
            name = id
            if project is None:
                raise TypeError(
                    'project has to be specified if instantiated' +
                    f'from track name. Track name: "{name}".'
                )
            track = project._get_track_by_name(name)
            if track is None:
                raise KeyError(f'track with name "{name}" can not be found.')
            id = track.id
            if id.endswith("0x0000000000000000"):
                raise KeyError(name)
            self._project = project
        # id is now a real ReaScript ID
        self.id = id

    @property
    def _args(self) -> ty.Tuple[str]:
        return self.id,

    @classmethod
    def _get_id_from_pointer(cls, pointer: ty.Union[float, str]) -> str:
        return '(MediaTrack*)0x{0:0{1}X}'.format(int(pointer), 16)

    @reapy_boost.inside_reaper()
    def _get_project(self) -> Project:
        """
        Return parent project of track.

        Should only be used internally; one should directly access
        Track.project instead of calling this method.
        """
        for project in reapy_boost.get_projects():
            if self.id in [t.id for t in project.tracks]:
                return project
        raise KeyError('Project can not be found. Maybe track is deleted.')

    def add_audio_accessor(self) -> AudioAccessor:
        """
        Create audio accessor and return it.

        Returns
        -------
        audio_accessor : AudioAccessor
            Audio accessor on track.
        """
        audio_accessor_id = RPR.CreateTrackAudioAccessor(  # type:ignore
            self.id)
        audio_accessor = reapy_boost.AudioAccessor(audio_accessor_id)
        return audio_accessor

    def add_fx(
        self,
        name: str,
        input_fx: bool = False,
        even_if_exists: bool = True
    ) -> FX:
        """
        Add FX to track and return it.

        Parameters
        ----------
        name : str
            FX name.
        input_fx : bool, optional
            Whether the FX should be an input (aka recording) FX or a
            standard FX (default=False). Note that if the track is the
            master track, input_fx=True will create a monitoring FX.
        even_if_exists : bool, optional
            Whether the FX should be added even if there already is an
            instance of the same FX on the track (default=True).

        Returns
        -------
        fx : FX
            New FX on track (or previously existing instance of FX if
            even_if_exists=False).

        Raises
        ------
        ValueError
            If there is no FX with the specified name.
        """
        index = RPR.TrackFX_AddByName(  # type:ignore
            self.id, name, input_fx, 1 - 2 * even_if_exists)
        if index == -1:
            raise ValueError("Can't find FX named {}".format(name))
        fx = reapy_boost.FX(ty.cast(reapy_boost.Track, self), index)
        return fx

    @reapy_boost.inside_reaper()
    def add_item(
        self,
        start: float = 0,
        end: ty.Optional[float] = None,
        length: float = 0
    ) -> Item:
        """
        Create new item on track and return it.

        Parameters
        ----------
        start : float, optional
            New item start in seconds (default=0).
        end : float, optional
            New item end in seconds (default None). If None, `length`
            is used instead.
        length : float, optional
            New item length in seconds (default 0).

        Returns
        -------
        item : Item
            New item on track.
        """
        if end is None:
            end = start + length
        item = reapy_boost.Item(
            RPR.AddMediaItemToTrack(self.id)
        )  # type:ignore
        item.position = start
        item.length = end - start
        return item

    def add_midi_item(
        self,
        start: float = 0,
        end: float = 1,
        quantize: bool = False
    ) -> Item:
        """
        Add empty MIDI item to track and return it.

        Parameters
        ----------
        start : float, optional
            New item start in seconds (or beats if `quantize`=True).
        end : float, optional
            New item end in seconds (or beats if `quantize`=True).
        quantize : bool, optional
            Whether to count time in beats (True) or seconds (False,
            default).
        """
        item_id = RPR.CreateNewMIDIItemInProj(  # type:ignore
            self.id, start, end, quantize)[0]
        item = reapy_boost.Item(item_id)
        return item

    def add_send(self, destination: ty.Optional['Track'] = None):
        """
        Add send to track and return it.

        Parameters
        ----------
        destination : Track or None
            Send destination (default=None). If None, destination is
            set to hardware output.

        Returns
        -------
        send : Send
            New send on track.
        """
        if isinstance(destination, Track):
            destination_id = destination.id
        send_id = RPR.CreateTrackSend(self.id, destination_id)  # type:ignore
        type = "hardware" if destination is None else "send"
        send = reapy_boost.Send(
            ty.cast(reapy_boost.Track, self), send_id, type=type
        )
        return send

    @property
    def automation_mode(self) -> str:  # TODO: make Enum for modes?
        """
        Track automation mode.

        One of the following values:
            "latch"
            "latch preview"
            "read"
            "touch"
            "trim/read"
            "write"

        :type: str
        """
        modes = "trim/read", "read", "touch", "write", "latch", "latch preview"
        automation_mode = modes[
            RPR.GetTrackAutomationMode(  # type:ignore
                self.id
            )]
        return automation_mode

    @automation_mode.setter
    def automation_mode(self, mode: str) -> None:
        """
        Set track automation mode.

        Parameters
        -------
        mode : str
            One of the following values:
                "latch"
                "latch preview"
                "read"
                "touch"
                "trim/read"
                "write"
        """
        modes = "trim/read", "read", "touch", "write", "latch", "latch preview"
        RPR.SetTrackAutomationMode(  # type:ignore
            self.id, modes.index(mode)
        )

    @property
    def color(self) -> ty.Tuple[int, int, int]:  # TODO: make color class?
        """
        Track color in RGB format.

        :type: tuple of int
        """
        native_color = RPR.GetTrackColor(self.id)  # type:ignore
        r, g, b = reapy_boost.rgb_from_native(native_color)
        return r, g, b

    @color.setter
    def color(self, color: ty.Tuple[int, int, int]) -> None:
        """
        Set track color to `color`

        Parameters
        ----------
        color : tuple
            Triplet of integers between 0 and 255 corresponding to RGB
            values.
        """
        native_color = reapy_boost.rgb_to_native(color)
        RPR.SetTrackColor(self.id, native_color)  # type:ignore

    def delete(self) -> None:
        """
        Delete track.
        """
        # TODO: can we verify?
        RPR.DeleteTrack(self.id)  # type:ignore

    @property
    def depth(self) -> int:
        """
        Track depth.

        :type: int
        """
        depth = RPR.GetTrackDepth(self.id)  # type:ignore
        return depth

    @property
    def envelopes(self) -> 'reapy_boost.EnvelopeList':
        """
        List of envelopes on track.

        :type: EnvelopeList
        """
        return reapy_boost.EnvelopeList(ty.cast(reapy_boost.Track, self))

    @classmethod
    def from_GUID(
        cls,
        guid_string: str,
        project: ty.Optional[Project] = None
    ) -> 'reapy_boost.Track':
        """
        Get track from project by GUID string {xyz-...}.

        Parameters
        ----------
        guid_string : str
            can be taken by Track.GUID
        project : Union[Project, str], optional
            if None — current project is used
            if "all" — iterates through all projects (use with care)

        Returns
        -------
        Track

        Raises
        ------
        KeyError
            If no track with the guid string found in project
        """
        if isinstance(project, str):
            if project != "all":
                raise TypeError(
                    'if used string argument — only "all" is accepted'
                )
            return cls._from_GUID_all_projects(guid_string)
        proj = project if project is not None else reapy_boost.Project()
        return proj.get_track_by_guid(guid_string)

    @classmethod
    def _from_GUID_all_projects(cls, guid_string: str) -> 'reapy_boost.Track':
        with reapy_boost.inside_reaper():
            for pr in reapy_boost.get_projects():
                try:
                    return pr.get_track_by_guid(guid_string)
                except (reapy_boost.errors.DistError, KeyError) as e:
                    if str(e).find('KeyError: {'):
                        continue
                    raise e
        raise KeyError(guid_string)

    @property
    def fxs(self) -> 'reapy_boost.FXList':
        """
        List of FXs on track.

        :type: FXList
        """
        fxs = reapy_boost.FXList(ty.cast(reapy_boost.Track, self))
        return fxs

    def get_info_string(self, param_name: str) -> str:
        return RPR.GetSetMediaTrackInfo_String(  # type:ignore
            self.id, param_name, "", False)[3]

    def get_info_value(self, param_name: str) -> float:
        value = RPR.GetMediaTrackInfo_Value(  # type:ignore
            self.id, param_name)
        return value

    @property
    def GUID(self) -> str:
        """
        Track's GUID string {xyz-...}.

        :type: str
        """
        return self.get_info_string("GUID")

    @GUID.setter
    def GUID(self, guid_string: str) -> None:
        self.set_info_string("GUID", guid_string)

    @property
    def has_valid_id(self):
        """
        Whether ReaScript ID is still valid.

        For instance, if track has been deleted, ID will not be valid
        anymore.

        :type: bool
        """
        return self._has_valid_id_inside()

    @reapy_boost.inside_reaper()
    def _has_valid_id_inside(self) -> bool:
        pointer, name = self._get_pointer_and_name()
        if self._project is None:
            return any(
                RPR.ValidatePtr2(p.id, pointer, name)  # type:ignore
                for p in reapy_boost.get_projects()
            )
        return bool(
            RPR.ValidatePtr2(  # type:ignore
                self.project.id, pointer, name
            )
        )

    @property
    def icon(self) -> str:
        """
        Track icon.

        Full filename, or relative to resource_path/data/track_icons.

        :type: str
        """
        return self.get_info_string("P_ICON")

    @icon.setter
    def icon(self, filename: str) -> None:
        self.set_info_string("P_ICON", filename)

    @property
    def index(self) -> ty.Optional[int]:
        """Track index in GUI (0-based).

        Will be ``None`` for master track.

        :type: int or None

        Raises
        ------
        InvalidObjectError
            When track does not exist in REAPER.
        """
        index = int(self.get_info_value('IP_TRACKNUMBER')) - 1
        if index >= 0:
            return index
        if index == -1:
            raise InvalidObjectError(self)

    @property
    def instrument(self) -> ty.Optional['reapy_boost.FX']:
        """
        First instrument FX on track if it exists, else None.

        :type: FX or None
        """
        fx_index = RPR.TrackFX_GetInstrument(self.id)  # type:ignore
        instrument = None if fx_index == -1 else reapy_boost.FX(self, fx_index)
        return instrument

    @property
    def items(self) -> ty.List['reapy_boost.Item']:
        """
        List of items on track.

        :type: list of Item
        """
        return self._items_inside()

    @reapy_boost.inside_reaper()
    def _items_inside(self) -> ty.List['reapy_boost.Item']:
        n_items = RPR.CountTrackMediaItems(self.id)  # type:ignore
        item_ids = [
            RPR.GetTrackMediaItem(  # type:ignore
                self.id, i
            ) for i in range(n_items)
        ]
        return list(map(reapy_boost.Item, item_ids))

    @property
    def is_muted(self) -> bool:
        """
        Whether track is muted.

        Can be manually set to change track state.
        """
        return bool(self.get_info_value("B_MUTE"))

    @is_muted.setter
    def is_muted(self, muted: bool) -> None:
        if muted:
            self.mute()
        else:
            self.unmute()

    @property
    def is_selected(self) -> bool:
        """
        Whether track is selected.

        :type: bool
        """
        is_selected = bool(RPR.IsTrackSelected(  # type:ignore
            self.id
        ))
        return is_selected

    @is_selected.setter
    def is_selected(self, selected: bool) -> None:
        """
        Select or unselect track.

        Parameters
        ----------
        selected : bool
            Whether to select or unselect track.
        """
        if selected:
            self.select()
        else:
            self.unselect()

    @property
    def is_solo(self) -> bool:
        """
        Whether track is solo.

        Can be manually set to change track state.
        """
        return bool(self.get_info_value("I_SOLO"))

    @is_solo.setter
    def is_solo(self, solo: bool) -> None:
        if solo:
            self.solo()
        else:
            self.unsolo()

    def make_only_selected_track(self) -> None:
        """
        Make track the only selected track in parent project.
        """
        RPR.SetOnlyTrackSelected(self.id)  # type:ignore

    def midi_hash(self, notes_only: bool = False) -> str:
        """Get hash of MIDI-data to compare with later.

        Parameters
        ----------
        notes_only : bool, (False by default)
            count only notes if True

        Returns
        -------
        str
        """
        return RPR.MIDI_GetTrackHash(  # type:ignore
            self.id, notes_only, 'hash', 1024**2)[3]

    @property
    def midi_note_names(self) -> ty.List[str]:
        with reapy_boost.inside_reaper():
            names = [
                RPR.GetTrackMIDINoteName(  # type:ignore
                    self.index, i, 0
                ) for i in range(128)
            ]
            return names
        raise InsideREAPERError('Contex manager exited without value.')

    @reapy_boost.inside_reaper()
    def mute(self) -> None:
        """Mute track (do nothing if track is already muted)."""
        if not self.is_muted:
            self.toggle_mute()

    @property
    def n_envelopes(self) -> int:
        """
        Number of envelopes on track.

        :type: int
        """
        n_envelopes = RPR.CountTrackEnvelopes(self.id)  # type:ignore
        return n_envelopes

    @property
    def n_fxs(self) -> int:
        """
        Number of FXs on track.

        :type: int
        """
        n_fxs = RPR.TrackFX_GetCount(self.id)  # type:ignore
        return n_fxs

    @property
    def n_hardware_sends(self) -> int:
        """
        Number of hardware sends on track.

        :type: int
        """
        n_hardware_sends = RPR.GetTrackNumSends(self.id, 1)  # type:ignore
        return n_hardware_sends

    @property
    def n_items(self) -> int:
        """
        Number of items on track.

        :type: int
        """
        n_items = RPR.CountTrackMediaItems(self.id)  # type:ignore
        return n_items

    @property
    def n_receives(self) -> int:
        n_receives = RPR.GetTrackNumSends(self.id, -1)  # type:ignore
        return n_receives

    @property
    def n_sends(self) -> int:
        n_sends = RPR.GetTrackNumSends(self.id, 0)  # type:ignore
        return n_sends

    @property
    def name(self) -> str:
        """
        Track name.

        Name is "MASTER" for master track, "Track N" if track has no
        name.

        :type: str
            Track name .
        """
        _, _, name, _ = RPR.GetTrackName(self.id, "", 2048)  # type:ignore
        return name

    @name.setter
    def name(self, track_name: str) -> None:
        self.set_info_string("P_NAME", track_name)

    @property
    def parent_track(self) -> ty.Optional['reapy_boost.Track']:
        """
        Parent track, or None if track has none.

        :type: Track or NoneType
        """
        parent = reapy_boost.Track(RPR.GetParentTrack(self.id))  # type:ignore
        if not parent._is_defined:
            parent = None
        return parent

    @property
    def project(self) -> 'reapy_boost.Project':
        """
        Track parent project.

        :type: Project
        """
        if self._project is None:
            self._project = self._get_project()
        return self._project

    @property
    def receives(self):
        return self._recieves_inside()

    @reapy_boost.inside_reaper()
    def _recieves_inside(self) -> ty.List['reapy_boost.Send']:
        return [
            reapy_boost.Send(
                ty.cast(reapy_boost.Track, self), i, type="receive"
            ) for i in range(self.n_receives)
        ]

    def select(self):
        """
        Select track.
        """
        RPR.SetTrackSelected(self.id, True)  # type:ignore

    @property
    def sends(self):
        return self._send_inside()

    @reapy_boost.inside_reaper()
    def _send_inside(self) -> ty.List['reapy_boost.Send']:
        return [
            reapy_boost.Send(reapy_boost.Track(self.id), i, type="send")
            for i in range(self.n_sends)
        ]

    def set_info_string(self, param_name: str, param_string: str) -> None:
        RPR.GetSetMediaTrackInfo_String(  # type:ignore
            self.id, param_name, param_string, True)

    def set_info_value(self, param_name: str, param_value: str) -> None:
        RPR.SetMediaTrackInfo_Value(  # type:ignore
            self.id, param_name, param_value)

    @reapy_boost.inside_reaper()
    def solo(self) -> None:
        """Solo track (do nothing if track is already solo)."""
        if not self.is_solo:
            self.toggle_solo()

    @reapy_boost.inside_reaper()
    def toggle_mute(self) -> None:
        """Toggle mute on track."""
        selected_tracks = self.project.selected_tracks
        self.make_only_selected_track()
        self.project.perform_action(40280)
        self.project.selected_tracks = selected_tracks

    @reapy_boost.inside_reaper()
    def toggle_solo(self) -> None:
        """Toggle solo on track."""
        selected_tracks = self.project.selected_tracks
        self.make_only_selected_track()
        self.project.perform_action(7)
        self.project.selected_tracks = selected_tracks

    @reapy_boost.inside_reaper()
    def unmute(self) -> None:
        """Unmute track (do nothing if track is not muted)."""
        if self.is_muted:
            self.toggle_mute()

    def unselect(self) -> None:
        """
        Unselect track.
        """
        RPR.SetTrackSelected(self.id, False)  # type:ignore

    @reapy_boost.inside_reaper()
    def unsolo(self) -> None:
        """Unsolo track (do nothing if track is not solo)."""
        if self.is_solo:
            self.toggle_solo()

    @property
    def visible_fx(self) -> ty.Optional['reapy_boost.FX']:
        """
        Visible FX in FX chain if any, else None.

        :type: FX or NoneType
        """
        with reapy_boost.inside_reaper():
            return self.fxs[
                RPR.TrackFX_GetChainVisible(  # type:ignore
                    self.id
                )]


TrackListKeyType = ty.Union[slice, int]


class TrackList(ReapyObjectList):
    """
    Container for a project's track list.

    Examples
    --------
    >>> tracks = project.tracks
    >>> len(tracks)
    4
    >>> tracks[0].name
    'Kick'
    >>> for track in tracks:
    ...     print(track.name)
    ...
    'Kick'
    'Snare'
    'Hi-hat'
    'Cymbal"
    """

    def __init__(self, parent: 'reapy_boost.Project') -> None:
        """
        Create track list.

        Parameters
        ----------
        parent : Project
            Parent project.
        """
        self.parent = parent

    @reapy_boost.inside_reaper()
    def __delitem__(self, key: TrackListKeyType) -> None:
        tracks = self[key] if isinstance(key, slice) else [self[key]]
        for track in tracks:
            track.delete()

    @ty.overload
    def __getitem__(self, key: slice) -> ty.List['reapy_boost.Track']:
        ...

    @ty.overload
    def __getitem__(self, key: int) -> 'reapy_boost.Track':
        ...

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self._get_items_from_slice(key)
        return reapy_boost.Track(key, self.parent)

    def __iter__(self):
        tracks = self[:]  # Only cost one distant call
        for track in tracks:
            yield track

    def __len__(self):
        return self.parent.n_tracks

    @property
    def _args(self):
        return self.parent,

    @reapy_boost.inside_reaper()
    def _get_items_from_slice(self,
                              slice: slice) -> ty.List['reapy_boost.Track']:
        indices = range(*slice.indices(len(self)))
        return [self[i] for i in indices]
        # return [self[i] for i in indices]



"""Get setup infos."""

from typing import Tuple
from reapy_boost.reascripts import enable_dist_api, disable_dist_api

import os
import sys


def get_config_scripts() -> Tuple[str, str]:
    """
    Return paths to configuration ReaScripts.
    """
    return (
        os.path.abspath(enable_dist_api.__file__),
        os.path.abspath(disable_dist_api.__file__)
    )


def get_python_dll() -> str:
    """
    Return path to Python DLL (if it can be found).
    """
    dir = os.path.dirname(sys.executable)
    file = os.path.basename(dir).lower() + ".dll"
    path = os.path.join(dir, file)
    if os.path.isfile(path):
        return path
    else:
        raise FileNotFoundError("Can't find python DLL...")


string = """
======================
  reapy_boost config infos
======================

Python DLL
----------
    {}

Enable or disable reapy_boost dist API
--------------------------------
Enable dist API
    {}

Disable dist API
    {}
"""

if __name__ == "__main__":
    try:
        dll = get_python_dll()
    except FileNotFoundError as e:
        dll = e.args[0]
    enable, disable = get_config_scripts()
    print(string.format(dll, enable, disable))



import re
import textwrap
import typing as ty
import builtins
import json
import platform as plt
import os
import warnings
import requests
import typing_extensions as te


def get_header_contents() -> str:
    """
    Get js_ReaScriptAPI header from GitHub as raw string.

    Returns
    -------
    str
    """
    root_raw = 'https://raw.githubusercontent.com/'
    header = (
        'juliansader/ReaExtensions/master/js_ReaScriptAPI/' +
        'Source%20code/js_ReaScriptAPI_def.h'
    )
    r = requests.get("{}{}".format(root_raw, header))
    return r.content.decode()


def write_binary_file(target_dir: str) -> None:
    """
    Get binary extension file from GitHub.

    Parameters
    ----------
    target_dir : str
        target directory path, usually:
        "{}{}{}".format(reapy_boost.get_resource_path, os.sep, UserPlugins)
    """
    root_api = 'https://api.github.com/repos/'
    repo = 'juliansader/ReaExtensions/contents/js_ReaScriptAPI/'
    url = "%s%s" % (root_api, repo)
    r = requests.get(url)
    tree = json.loads(r.content)
    name = ''

    for item in tree:
        if item['name'] > name:
            name = item['name']
            bindir = item
    r = requests.get(bindir['_links']['self'])

    tree = json.loads(r.content)

    ext_need = {
        ('Linux', '64bit'): '.so',
        ('Windows', '64bit'): '64.dll',
        ('Windows', '32bit'): '32.dll',
        ('Darwin', '64bit'): '64.dylib',
        ('Darwin', '32bit'): '32.dylib',
    }
    ext_real = plt.system(), plt.architecture()[0]
    bin_url, bin_name = '', ''
    for item in tree:
        if item['name'].endswith(ext_need[ext_real]):
            bin_url = item['download_url']
            bin_name = item['name']
    if not bin_url:
        warnings.warn(
            'binary release not found:' +
            ' sometimes Julian forgets about your platform.' +
            "don't worry and install it via ReaPack"
        )
        return
    r = requests.get(bin_url)

    with open(
        os.path.normpath("%s%s%s" % (target_dir, os.sep, bin_name)), 'wb'
    ) as f:
        f.write(r.content)


ArgsT = ty.Dict[str, str]

FuncdefT = te.TypedDict(
    'FuncdefT', {
        'name': str,
        'return': str,
        'doc': str,
        'args': ArgsT
    }
)


class Parcer:
    """Parces cpp header of js_ReaScriptExtensions.

    Attributes
    ----------
    defs : List[FuncdefT]
        list of parced function defs
    """

    def __init__(
        self,
        file: ty.Optional[str] = None,
        raw_str: ty.Optional[str] = None
    ) -> None:
        self._re_line = re.compile(r'(?!//).*{ APIFUNC\(.+')
        self._re_parce_line = re.compile(r'".*?"')
        self._re_sub = re.compile(r'[\s\{\}]')
        self._re_sub_q = re.compile(r'\\"')
        self._re_name = re.compile(r'.*APIFUNC\((.+?)\)')
        self.types_str: ty.Set[str] = set()
        if file is not None:
            with open(file) as f:
                self.defs = self._parce(f.readlines())
        if raw_str is not None:
            self.defs = self._parce(raw_str.splitlines())

    def _parce(self, lines: ty.List[str]) -> ty.List[FuncdefT]:
        defs = []
        for line in lines:
            line = line.strip()
            if re.match(self._re_line, line):
                defs.append(self._parce_line(line))
        return defs

    def _parce_line(self, line: str) -> FuncdefT:
        line = self._re_sub_q.sub("'", line)
        m = self._re_parce_line.findall(line)
        ret, types, names, doc = m
        doc = re.sub(r'\\n', '\n', doc)
        out: FuncdefT = {}  # type:ignore
        name = self._re_name.match(line).group(1)  # type:ignore

        out['name'] = name
        out['return'] = ret[1:-1]
        out['args'] = {}
        self.types_str.add(out['return'])
        for type_, name in zip(types[1:-1].split(','), names[1:-1].split(',')):
            if type_ == '':
                continue
            type_ = type_.replace('const ', '')
            name = re.sub(r'[\(\)]', '', name)
            if name in builtins.__dict__:
                name += '_'
            out['args'][name] = type_
            self.types_str.add(type_)
        out['doc'] = doc[1:-1]
        # if name == "JS_Window_GetRect":
        # print(out)

        # print(out)
        return out


TypeDefT = te.TypedDict(
    'TypeDefT', {
        'def': str,
        'ref': str,
        'prot': str,
        'call': str,
        'ret': str,
        'default': str,
    }
)

_types_to_match: ty.Dict[str, TypeDefT] = {
    'ptr':
        {
            'def': '{type_}',
            'ref': 'ct.c_uint64',
            'prot': '',
            'call': 'packp("{type_}", str({arg}))',
            'ret': 'Pointer(unpackp("{type_}", ret), "{type_}")',
            'default': '',
        },
    'void*':
        {
            'def': 'VoidPtr',
            # 'ref': 'ct.c_uint64',
            'ref': 'ct.c_void_p',
            'prot': '',
            'call': 'packp("void*", str({arg}))',
            'ret': 'VoidPtr(unpackp("void*", ret))',
            'default': '',
        },
    'void':  # noob formatting comment
        {
            'def': 'None',
            'ref': 'None',
            'prot': '',
            'call': '',
            'ret': '',
            'default': '',
        },
    'double*':
        {
            'def': 'float',
            'ref': 'ct.c_void_p',
            'prot': 'prot_{arg} = ct.c_double({val})',
            'call': 'ct.byref(prot_{arg})',
            'ret': 'float(prot_{arg}.value)',
            'default': '1.0',
        },
    'int*':
        {
            'def': 'int',
            'ref': 'ct.c_void_p',
            'prot': 'prot_{arg} = ct.c_int({val})',
            'call': 'ct.byref(prot_{arg})',
            'ret': 'prot_{arg}.value',
            'default': '1',
        },
    'bool*':
        {
            'def': 'bool',
            'ref': 'ct.c_void_p',
            'prot': 'prot_{arg} = ct.c_byte({val})',
            'call': 'ct.byref(prot_{arg})',
            'ret': 'bool(prot_{arg}.value)',
            'default': 'False',
        },
    'char*':
        {
            'def': 'str',
            'ref': 'ct.c_char_p',
            'prot':
                ('prot_{arg} = packs_l({arg}, size=len({arg}.encode(encoding))'
                    ', encoding=encoding)'),
            'call': 'prot_{arg}',
            'ret': 'unpacks_l(prot_{arg}, want_raw=want_raw)',
            'default': 'prot_{arg} = packs_l("", size={size}, encoding=encoding)',
        },
    'double':
        {
            'def': 'float',
            'ref': 'ct.c_double',
            'prot': '',
            'call': 'ct.c_double({arg})',
            'ret': 'float({arg})',
            'default': '',
        },
    'int':
        {
            'def': 'int',
            'ref': 'ct.c_int',
            'prot': '',
            'call': 'ct.c_int({arg})',
            'ret': 'int({arg})',
            'default': '',
        },
    'bool':
        {
            'def': 'bool',
            'ref': 'ct.c_byte',
            'prot': '',
            'call': 'ct.c_byte({arg})',
            'ret': 'bool({arg})',
            'default': '',
        },
    'char':
        {
            'def': 'bytes',
            'ref': 'ct.c_char',
            'prot': '',
            'call': 'ct.c_char({arg})',
            'ret': 'bytes({arg})',
            'default': '',
        },
}

_def_templ = """
@reapy_boost.inside_reaper()
def {def_name}({def_args}) -> {def_ret}:
    \"""
    {doc}

    Returns
    -------
    {doc_ret}
    \"""
    a = _ft['{api_name}']
    f = ct.CFUNCTYPE({ref_ret}, {ref_args})(a)
    {prot_args}
    ret = f({call_args})
    return {ret_full}
"""

_class_def_templ = """
class {type_}(Pointer):
    pass
"""

_module_templ = \
    """\"\"\"Autogenerated API for ReaExtensions by Julian Sader.

Repo on GitHub: https://github.com/juliansader/ReaExtensions
Theme on ReaperForums: https://forum.cockos.com/showthread.php?t=212174
\"\"\"

import typing as ty
import reapy_boost
import ctypes as ct
from reapy_boost.core import ReapyObject
if reapy_boost.is_inside_reaper():
    from reapy_boost.additional_api import packp, unpackp, packs_l, unpacks_l
    from reaper_python import _ft
else:
    from reapy_boost.additional_api import packp, unpackp

MAX_STRBUF = 4 * 1024 * 1024

__all__: ty.List[str] = [
    {all_}
]


class Pointer(ReapyObject):

    def __init__(
        self, ptr: ty.Union[str, int], ptr_str: str = "void*"
    ) -> None:
        self.__args = ptr, ptr_str
        if isinstance(ptr, str):
            self._str = ptr
            self._int = packp(ptr_str, ptr)
            return
        if isinstance(ptr, int):
            self._str = unpackp(ptr_str, ptr)
            self._int = ptr
            return
        raise TypeError("expect int or str, passed %s" % ptr)

    @property
    def _args(self) -> ty.Tuple[ty.Union[str, int], str]:
        return self.__args

    def __str__(self) -> str:
        return self._str

    def __int__(self) -> int:
        return self._int

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (str, Pointer)):
            return self._str == other
        if isinstance(other, int):
            return self._int == other
        return False

    def __hash__(self) -> int:
        return int(self)


class VoidPtr(Pointer):
    pass

{class_defs}

{defs}
"""


class FuncBuilder:
    """Builds API as python code.

    Attributes
    ----------
    matched_types : Dict[str, TypeDefT]
        matched ctypes to python types, considering their definitions
    parcer : Parcer
    """

    def __init__(self, parcer: Parcer) -> None:
        self.parcer = parcer
        self._re_ptr = re.compile(r'[A-Z].+\*')
        self.all: ty.List[str] = ["Pointer", "VoidPtr"]
        self.matched_types = self._match_types()

    def _match_types(self) -> ty.Dict[str, TypeDefT]:
        out: ty.Dict[str, TypeDefT] = {}
        for type_ in self.parcer.types_str:
            if type_ in _types_to_match:
                out[type_] = _types_to_match[type_]
                continue
            if self._re_ptr.match(type_):
                out[type_] = _types_to_match['ptr']
                self.all.append(type_[:-1])
                continue
            raise TypeError('Cannot match {}'.format(type_))
        return out

    def _match_def_args(
        self, args: ty.Dict[str, str], out_names: ty.List[str],
        opt_names: ty.List[str], opt_defs: ty.List[str]
    ) -> str:
        defs = {}
        kwargs = {}
        types = set()
        for a, t in args.items():
            matched_t = self.matched_types[t]['def']
            types.add(matched_t)
            if a in out_names:
                if matched_t == 'str':
                    kwargs['want_raw'] = 'bool=False'
                continue
            if a in opt_names:
                odct = opt_defs.pop(0).split(':')
                # print(opt_defs_c)
                defs[odct[0]] = odct[1]
                continue
            defs[a] = matched_t.format(type_=t[:-1])
        if 'str' in types:
            kwargs['encoding'] = 'str="utf-8"'
        defs.update(kwargs)
        return ', '.join(
            '{arg}: {type_}'.format(arg=a, type_=t) for a, t in defs.items()
        )

    def _match_def_ret(
        self, ret: str, args: ty.Dict[str, str], out_names: ty.List[str]
    ) -> str:
        base = self.matched_types[ret]['def'].format(type_=ret).strip('*')
        if not out_names:
            return base
        total_names = out_names
        if base == 'None' and len(total_names) == 1:
            return self.matched_types[args[total_names[0]]]['def']
        if base != 'None':
            out = 'ty.Tuple[{ret}, {args}]'
        else:
            out = 'ty.Tuple[{args}]'
        return out.format(
            ret=base.format(type_=ret[:-1]),
            args=', '.join(
                self.matched_types[args[n]]['def'].format(type_=args[n][:-1])
                for n in total_names
            )
        )

    def _match_call_args(
        self, args: ty.Dict[str, str], opt_names: ty.List[str],
        out_names: ty.List[str], opt_calls: ty.List[str],
        out_calls: ty.List[str]
    ) -> ty.Tuple[ty.List[str], str]:
        """
        Returns
        -------
        Tuple[List[str], str]
            prototype defs, calls line
        """
        prots, calls = [], []

        for a, t in args.items():
            if a not in [*opt_names, *out_names]:
                matched_call = self.matched_types[t]['call']
                calls.append(matched_call.format(type_=t, arg=a))
                if self.matched_types[t]['def'] == 'str':
                    prots.append(
                        self.matched_types[t]['prot'].format(arg=a, type_=t)
                    )
                continue
            elif a in opt_names:
                calls.append(opt_calls.pop(0))
                continue
            elif a in out_names:
                calls.append(out_calls.pop(0))
                continue
            raise KeyError(a)
        return prots, ', '.join(calls)

    def _match_optional_args(
        self, args: ArgsT
    ) -> ty.Tuple[ty.List[str], ty.List[str], ty.List[str], ty.List[str]]:
        """
        Returns
        -------
        Tuple[List[str], List[str], List[str], List[str]]
            names, defs, prots, calls
        """
        defs, calls, prots, names = [], [], [], []
        def_t = "{arg}: ty.Optional[{type_}]=None"
        call_t = "None if {arg} is None else {call}"
        prot_t = "{prot} if {arg} is not None else None"
        for a, t in args.items():
            if a.find('Optional') == -1:
                continue
            names.append(a)
            a = a.replace('InOptional', '')
            m_def = self.matched_types[t]['def'].format(type_=t)
            defs.append(def_t.format(arg=a, type_=m_def))

            m_prot = self.matched_types[t]['prot'].format(
                arg=a, val=a, type_=t
            )
            if m_prot:
                prots.append(prot_t.format(prot=m_prot, arg=a))

            m_call = self.matched_types[t]['call'].format(arg=a, type_=t)
            calls.append(call_t.format(arg=a, call=m_call))
        return names, defs, prots, calls

    def _match_out_args(
        self, args: ArgsT
    ) -> ty.Tuple[ty.List[str], ty.List[str], ty.List[str], ty.List[str],
                  ty.List[str]]:
        """
        Returns
        -------
        Tuple[ty.List[str], ty.List[str], ty.List[str], ty.List[str],
            ty.List[str]]
            names, prots, calls, rets, docs
        """
        rets, calls, prots, names, docs = [], [], [], [], []
        for a, t in args.items():
            if a.find('Out') == -1 or not t.endswith('*'):
                continue
            if a.endswith('_sz'):
                # still, only three size args supports passing byref
                # until the better times I'll leave this as it is
                args[a] = t.replace('*', '')
                continue
            names.append(a)
            a_orig = a
            a = a.replace('Out', '')
            if self.matched_types[t]['def'] == 'str':
                m_prot = self.matched_types[t]['default'].format(
                    arg=a, type_=t, size="{}_sz".format(a_orig)
                )
            else:
                m_prot = self.matched_types[t]['prot'].format(
                    arg=a, val=self.matched_types[t]['default'], type_=t
                )
            prots.append(m_prot)

            m_call = self.matched_types[t]['call'].format(arg=a, type_=t)
            calls.append(m_call)

            m_ret = self.matched_types[t]['ret'].format(arg=a, type_=t)
            rets.append(m_ret)
            docs.append(
                "{arg}: {type_}".format(
                    arg=a,
                    type_=self.matched_types[t]['def'].format(arg=a, type_=t)
                )
            )
        return names, prots, calls, rets, docs

    def build_def(self, item: FuncdefT) -> str:
        """
        Build the single def from parced funcdef.

        Parameters
        ----------
        item : FuncdefT

        Returns
        -------
        str
            multiline python definition
        """
        def_name = item['name'].replace('JS_', '')
        self.all.append(def_name)

        (opt_names, opt_defs, opt_prots,
         opt_calls) = self._match_optional_args(item['args'])
        (out_names, out_prots, out_calls, out_rets,
         out_docs) = self._match_out_args(item['args'])
        def_ret = self._match_def_ret(
            ret=item['return'], args=item['args'], out_names=out_names
        )
        doc = textwrap.fill(
            item['doc'],
            replace_whitespace=False,
        ).replace('\n', '\n    ')
        ret_def_str = self.matched_types[item['return']]['def']
        if ret_def_str == 'None':
            doc_ret_ret: ty.List[str] = []
        else:
            doc_ret_ret = ["ret_value: {}".format(ret_def_str)]
        doc_ret = '\n    '.join([*doc_ret_ret, *out_docs])
        api_name = item['name']
        ref_ret = self.matched_types[item['return']]['ref']
        ref_args = ', '.join(
            '{type_}'.format(type_=self.matched_types[t]['ref'])
            for t in item['args'].values()
        )
        norm_prots, call_args = self._match_call_args(
            item['args'], opt_names, out_names, opt_calls, out_calls
        )
        ret_ret = self.matched_types[item['return']]['ret'].format(
            type_=item['return'], arg='ret'
        )
        ret_full = ret_ret + (', ' if ret_ret and out_rets else '')
        ret_full += ', '.join(out_rets)
        prot_args = '\n    '.join([*opt_prots, *out_prots, *norm_prots])
        def_args = self._match_def_args(
            item['args'], out_names, opt_names, opt_defs
        )

        return _def_templ.format(
            def_name=def_name,
            def_args=def_args,
            def_ret=def_ret,
            doc=doc,
            doc_ret=doc_ret,
            api_name=api_name,
            ref_ret=ref_ret,
            ref_args=ref_args,
            prot_args=prot_args,
            call_args=call_args,
            ret_full=ret_full
        )

    def _match_pointer_types(self) -> ty.List[str]:
        out: ty.List[str] = []
        for t in self.matched_types.keys():
            if t not in _types_to_match:
                out.append(t[:-1])
        return out

    def build_module(self) -> str:
        """
        Build the whole module, based on parced defs.

        Returns
        -------
        str
            multiline str of the generated module
        """
        defs = self.parcer.defs
        defs_str = [self.build_def(def_) for def_ in defs]
        code = _module_templ.format(
            all_=',\n    '.join('"%s"' % s for s in self.all),
            class_defs='\n'.join(
                _class_def_templ.format(type_=t)
                for t in self._match_pointer_types()
            ),
            defs='\n'.join(defs_str)
        )

        # mybe some code formatter should be here
        return code


def generate_js_api(bin_dir: str, api_filename: str) -> None:
    """
    Get ReaExtensions and generate apropriate files.

    Parameters
    ----------
    bin_dir : str
        binary file directory usually:
        "{}{}{}".format(reapy_boost.get_resource_path, os.sep, UserPlugins)

    api_filepath : str
        file to be imported by end-user, 'reapy_boost/core/gui/JS_API.py' ?
    """
    header_raw = get_header_contents()
    write_binary_file(bin_dir)
    prc = Parcer(raw_str=header_raw)
    builder = FuncBuilder(prc)
    module_str = builder.build_module()
    dir_ = os.path.dirname(os.path.abspath(__file__))
    api_filepath = os.path.join(dir_, api_filename)
    with open(api_filepath, 'w') as f:
        f.write(module_str)
    js_filepath = os.path.join(dir_, 'JS_API.py')
    with open(js_filepath, 'r') as f:
        lines = []
        ignore = False
        for line in f.readlines():
            if ignore and ']' in line:
                line = '\n' + line
                ignore = False
            if not ignore:
                lines.append(line)
            if '__all__: ty.List[str] = [' in line:
                ignore = True
                lines.append(',\n    '.join('"%s"' % s for s in builder.all))
    with open(js_filepath, 'w') as f:
        f.writelines(lines)


if __name__ == '__main__':
    import reapy_boost
    api_filename = "_JS_API_generated.py"
    bin_dir = os.path.join(reapy_boost.get_resource_path(), "UserPlugins")
    # print(api_filepath)
    # print(bin_dir)
    generate_js_api(bin_dir, api_filename)



from types import TracebackType
from typing import Dict, List, Optional, Tuple
import reapy_boost
import reapy_boost.reascript_api as RPR
import contextlib
from .defer import ReaperConsole

import collections
import io
import os
import sys

_ORIGINAL_PRINT = print


def add_project_tab(
    make_current_project: bool = True
) -> 'reapy_boost.Project':
    """Open new project tab and return it.

    Parameters
    ----------
    make_current_project : bool
        Whether to select new project as current project
        (default=`True`).

    Returns
    -------
    project : Project
        New project.
    """
    if not make_current_project:
        current_project = reapy_boost.Project()
        project = add_project_tab(make_current_project=True)
        current_project.make_current_project()
        return project
    perform_action(40859)
    return reapy_boost.Project()


def add_reascript(path: str, section_id: int = 0, commit: bool = True) -> int:
    """
    Add a ReaScript and return the new action ID.

    Parameters
    ----------
    path : str
        Path to script.
    section_id : int, optional (default=0, corresponds to main section).
        Action section ID to which the script must be added.
    commit : bool, optional
        Whether to commit change. Use it when adding a single script.
        You can optimize bulk adding `n` scripts by setting
        `commit=False` for the first `n-1` calls and `commit=True` for
        the last call.

    Returns
    -------
    action_id : int
        New ReaScript action ID.
    """
    if not os.path.isfile(path):
        raise FileNotFoundError(path)
    path = os.path.abspath(path)
    action_id = RPR.AddRemoveReaScript(  # type:ignore
        True, section_id, path, commit)
    if action_id == 0:
        message = "Script at {} wasn't successfully added.".format(path)
        raise ValueError(message)
    return action_id


def arm_command(command_id: int, section: str = "") -> None:
    """
    Arm or disarm command.

    Parameters
    ----------
    command_id : int
        Command ID. If 0, disarm command.
    section : str, optional
        Command section. Empty string for main section. Default="".
        TODO: explain what section parameter corresponds to.
    """
    RPR.ArmCommand(command_id, section)  # type:ignore


def browse_for_file(window_title: str = "",
                    extension: str = "") -> Optional[str]:
    """
    Ask the user to select a file.

    Parameters
    ----------
    window_title : str, optional
        Window title (default="")
    extension : str, optional
        Extension for file (e.g. "mp3", "txt"...) (default=all types).

    Returns
    -------
    path : str or NoneType
        Path to file, or None if user cancelled.
    """
    success, path, *_ = RPR.GetUserFileNameForRead(  # type:ignore
        "", window_title, extension)
    if success:
        return path


def clear_console() -> None:
    """
    Clear Reaper console.

    See also
    --------
    ReaProject.show_console_message
    """
    RPR.ClearConsole()  # type:ignore


def clear_peak_cache() -> None:
    """
    Reset global peak cache.
    """
    RPR.ClearPeakCache()  # type:ignore


def dB_to_slider(db: float) -> float:
    """
    Convert decibel value to slider.

    Parameters
    ----------
    db : float
        Decibel value.

    Returns
    -------
    slider : float
        Slider value.

    See also
    --------
    slider_to_dB
    """
    slider = RPR.DB2SLIDER(db)  # type:ignore
    return slider


def delete_ext_state(section: str, key: str, persist: bool = False) -> None:
    """
    Delete extended state value for a given section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key.
    persist : bool
        Whether extended state should remain deleted next time REAPER
        is opened.
    """
    RPR.DeleteExtState(section, key, persist)  # type:ignore


def disarm_command() -> None:
    """
    Disarm command.
    """
    arm_command(0)


def get_armed_command() -> Optional[Tuple[int, str]]:
    command_id, section, _ = RPR.GetArmedCommand("", 2048)  # type:ignore
    if command_id == 0:
        return
    return command_id, section


def get_command_id(command_name: str) -> Optional[int]:
    """
    Return ID of command with a given name.

    Parameters
    ----------
    command_name : str
        Command name. Perhaps, id-string

    Returns
    -------
    command_id : int or None
        Command ID, or None if name can't be found.
    """
    command_id = RPR.NamedCommandLookup(command_name)  # type:ignore
    command_id = command_id if command_id else None
    return command_id


def get_command_name(command_id: int) -> str:
    """
    Return name of command with a given ID.

    Parameters
    ----------
    command_id : int
        Command ID.

    Returns
    -------
    command_name : str, None
        Command name, or None for a native command.
    """
    command_name = RPR.ReverseNamedCommandLookup(command_id)  # type:ignore
    if command_name is not None:
        command_name = "_" + command_name
    return command_name


def get_exe_dir() -> str:
    """
    Return REAPER.exe directory (e.g. "C:\\Program Files\\REAPER").

    Returns
    -------
    path : str
        Path to REAPER.exe directory.
    """
    path = RPR.GetExePath()  # type:ignore
    return path


def get_ext_state(section: str, key: str) -> str:
    """
    Get the extended state value for a specific section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key for section `section`.

    Returns
    -------
    value : str
        Extended state value.

    See also
    --------
    delete_ext_state
    set_ext_state
    """
    value = RPR.GetExtState(section, key)  # type:ignore
    return value


def get_global_automation_mode() -> str:
    """
    Return global automation override mode.

    Returns
    -------
    override_mode : str
        One of the following values:
            "bypass"
            "latch"
            "none"
            "read"
            "touch"
            "trim/read"
            "write"
    TODO: consider introducing of Enum
    """
    modes = {
        -1: "none",
        0: "trim/read",
        1: "read",
        2: "touch",
        3: "write",
        4: "latch",
        5: "bypass"
    }
    override_mode = modes[RPR.GetGlobalAutomationOverride()]  # type:ignore
    return override_mode


def get_ini_file() -> str:
    """
    Return path to REAPER.ini file.

    Returns
    -------
    path : str
        Path to REAPER.ini file.
    """
    path = RPR.get_ini_file()  # type:ignore
    return path


def get_last_touched_track() -> Optional['reapy_boost.Track']:
    """
    Return last touched track, or None if no track has been touched.

    Returns
    -------
    track : Track or None if no track has been touched.
    """
    track = reapy_boost.Track(RPR.GetLastTouchedTrack())  # type:ignore
    if not track._is_defined:
        track = None
    return track


def get_main_window() -> 'reapy_boost.Window':
    """
    Return main window.

    Returns
    -------
    window : Window
        Main window.
    """
    window = reapy_boost.Window(RPR.GetMainHwnd())  # type:ignore
    return window


@reapy_boost.inside_reaper()
def get_projects() -> List['reapy_boost.Project']:
    """
    Return list of all opened projects.

    Returns
    -------
    projects : list of Project
        List of all projects.
    """
    i, projects = 0, [reapy_boost.Project(index=0)]
    while projects[-1]._is_defined:
        i += 1
        projects.append(reapy_boost.Project(index=i))
    projects.pop()
    return projects


def get_reaper_version() -> str:
    version = RPR.GetAppVersion()  # type:ignore
    return version


def get_resource_path() -> str:
    """
    Return path to directory where .ini files are stored.

    Returns
    -------
    path : str
        Path to directory where .ini files are stored.
    """
    path = RPR.GetResourcePath()  # type:ignore
    return path


def get_user_inputs(title: str,
                    captions: List[str],
                    retvals_size: int = 1024) -> Dict[str, str]:
    """Show text inputs to user and get values from them.

    Parameters
    ----------
    title : str
        Popup title.
    captions : List[str]
        Names of input fields.
    retvals_size : int, optional
        Maximum number of characters that will be retrieved for each
        field. User may enter more, but only the first `retvals_size`
        will be returned. (default=1024)

    Returns
    -------
    Dict[str,str]
        Dictionary of pairs {caption: response}.

    Raises
    ------
    RuntimeError
        When user clicked the Cancel button.
    """
    success, _, _, _, retvals_csv, _ = RPR.GetUserInputs(  # type:ignore
        title, len(captions), ",".join(captions), "", retvals_size)
    if success:
        return dict(zip(captions, retvals_csv.split(",")))
    else:
        raise RuntimeError('User clicked Cancel.')


def has_ext_state(section: str, key: str) -> bool:
    """
    Return whether extended state exists for given section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key.

    Returns
    -------
    has_ext_state : bool
    """
    has_ext_state = bool(RPR.HasExtState(section, key))  # type:ignore
    return has_ext_state


def is_valid_id(id_: str) -> bool:
    """If it is not nullptr.

    Parameters
    ----------
    id_ : str

    Notes
    -----
    Not actually does anything else than id is not 0x00000

    Returns
    -------
    bool
    """
    return not id_.endswith("0x0000000000000000")


@reapy_boost.inside_reaper()
def open_project(
    filepath: str,
    in_new_tab: bool = False,
    make_current_project: bool = True
) -> 'reapy_boost.Project':
    """
    Open project and return it.

    Parameters
    ----------
    filepath : str
    in_new_tab : bool, optional
        Whether to open project in new tab (default=`False`).
    make_current_project : bool, optional
        Whether to make opened project current project (has no effect
        if `in_new_tab` is `False`).

    Returns
    -------
    project : Project
        Opened project.
    """
    if not make_current_project:
        current_project = reapy_boost.Project()
    if in_new_tab:
        add_project_tab(make_current_project=True)
    RPR.Main_openProject(filepath)  # type:ignore
    project = reapy_boost.Project()
    if not make_current_project:
        current_project.make_current_project()
    return project


def perform_action(action_id: int) -> None:
    """
    Perform action with ID `action_id` in the main Actions section.

    Parameters
    ----------
    action_id : int
        Action ID in the main Actions section.
    """
    RPR.Main_OnCommand(action_id, 0)  # type:ignore


class prevent_ui_refresh(contextlib.ContextDecorator):
    """Class to prevent UI refresh on certain pieces of code.

    Its instance can be used both as decorator and as context manager:

    >>> with reapy_boost.prevent_ui_refresh():
    ...     reapy_boost.Project.add_track()

    >>> @prevent_ui_refresh()
    >>> def some_function(*args, **kwargs):
    ...     reapy_boost.Project.add_track()

    """

    def __enter__(self) -> None:
        RPR.PreventUIRefresh(1)  # type:ignore

    def __exit__(
        self, exc_type: Exception, exc_val: str, exc_tb: TracebackType
    ) -> None:
        RPR.PreventUIRefresh(-1)  # type:ignore


def print(*args: object, **kwargs: object) -> None:
    """
    Alias to ReaProject.show_console_message.
    """
    show_console_message(*args, **kwargs)


class reaprint(contextlib.ContextDecorator):
    """Class to send all prints to ReaperConsole.

    Its instance can be used both as decorator and context manager:

    >>> with reapy_boost.reaprint():
    ...     print('This will go to the console!')
    ...     print('All these contexted will go to the console!')

    >>> @reapy_boost.reaprint()
    >>> def some_function(*args, **kwargs):
    ...     print('This will go to the console!')
    ...     print('All these decorated prints will go to the console!')

    """
    _original_stdouts = collections.deque()

    def __enter__(self) -> None:
        self._original_stdouts.append(sys.stdout)
        sys.stdout = ReaperConsole()

    def __exit__(
        self, exc_type: Exception, exc_val: str, exc_tb: TracebackType
    ) -> None:
        sys.stdout = self._original_stdouts.pop()


def remove_reascript(
    path: str, section_id: int = 0, commit: bool = True
) -> None:
    """
    Remove a ReaScript.

    Parameters
    ----------
    path : str
        Path to script.
    section_id : int, optional (default=0, corresponds to main section).
        Action section ID to which the script must be added.
    commit : bool, optional
        Whether to commit change. Use it when removing a single script.
        You can optimize bulk removing `n` scripts by setting
        `commit=False` for the first `n-1` calls and `commit=True` for
        the last call.
    """
    path = os.path.abspath(path)
    success = RPR.AddRemoveReaScript(  # type:ignore
        False, section_id, path, commit)
    if not success:
        message = "Script at {} wasn't successfully added.".format(path)
        raise ValueError(message)


def rgb_from_native(native_color: int) -> Tuple[int, int, int]:
    """
    Extract RGB values from a native (OS-dependent) color.

    Parameters
    ----------
    native_color : int
        Native color.

    Returns
    -------
    r, g, b : (int, int, int)
        RGB values between 0 and 255.
    """
    _, r, g, b = RPR.ColorFromNative(native_color, 0, 0, 0)  # type:ignore
    return r, g, b


def rgb_to_native(rgb: Tuple[int, int, int]) -> int:
    """
    Make a native (OS-dependent) color from RGB values.

    Parameters
    ----------
    rgb : (int, int, int)
        RGB triplet of integers between 0 and 255.

    Returns
    -------
    native_color : int
        Native color.
    """
    native_color = RPR.ColorToNative(*rgb)  # type:ignore
    return native_color


def set_ext_state(
    section: str, key: str, value: str, persist: bool = False
) -> None:
    """
    Set the extended state value for a specific section and key.

    Parameters
    ----------
    section : str
        Extended state section.
    key : str
        Extended state key for section `section`.
    value : str
        Extended state value for section `section` and key `key`.
    persist : bool
        Whether the value should be stored and reloaded the next time
        REAPER is opened.

    See also
    --------
    delete_ext_state
    get_ext_state
    """
    RPR.SetExtState(section, key, value, persist)  # type:ignore


def set_global_automation_mode(mode: str) -> None:
    """
    Set global automation mode.

    Parameters
    ----------
    mode : str
        One of the following values:
            "bypass"
            "latch"
            "none"
            "read"
            "touch"
            "trim/read"
            "write"
    TODO: consider introducing Enum
    """
    modes = {
        "none": -1,
        "trim/read": 0,
        "read": 1,
        "touch": 2,
        "write": 3,
        "latch": 4,
        "bypass": 5
    }
    RPR.SetGlobalAutomationOverride(modes[mode])  # type:ignore


def show_console_message(*args: object, sep: str = " ", end: str = "\n"):
    """
    Print a message to the Reaper console.

    Parameters
    ----------
    args : tuple
        Values to print.
    sep : str, optional
        String inserted between values (default=" ").
    end : str, optional
        String appended after the last value (default="\n").
    """
    file = io.StringIO()
    _ORIGINAL_PRINT(*args, sep=sep, end=end, file=file)
    file.seek(0)
    txt = file.read()
    RPR.ShowConsoleMsg(txt)  # type:ignore


def show_message_box(text: str = "", title: str = "", type: str = "ok"):
    """
    Show message box.

    Parameters
    ----------
    text : str
        Box message
    title : str
        Box title
    type : str
        One of the following values.

        "ok"
        "ok-cancel"
        "abort-retry-ignore"
        "yes-no-cancel"
        "yes-no"
        "retry-cancel"

    Returns
    -------
    status : str
        One of the following values.

        "ok"
        "cancel"
        "abort"
        "retry"
        "ignore"
        "yes"
        "no"
    """
    all_types = {
        "ok": 0,
        "ok-cancel": 1,
        "abort-retry-ignore": 2,
        "yes-no-cancel": 3,
        "yes-no": 4,
        "retry-cancel": 5
    }
    all_status = {
        1: "ok",
        2: "cancel",
        3: "abort",
        4: "retry",
        5: "ignore",
        6: "yes",
        7: "no"
    }
    status = RPR.ShowMessageBox(  # type:ignore
        text, title, all_types[type]
    )
    status = all_status[status]
    return status


def slider_to_dB(slider: float) -> float:
    """
    Convert slider value to decibel.

    Parameters
    ----------
    slider : float
        Slider value.

    Returns
    -------
    db : float
        Decibel value.

    See also
    --------
    dB_to_slider
    """
    db = RPR.SLIDER2DB(slider)  # type:ignore
    return db


def test_api() -> None:
    """Display a message window if the API can successfully be called."""
    RPR.APITest()  # type:ignore


class undo_block(contextlib.ContextDecorator):
    """Class to register undo block.

    Its instance can be used both as decorator and context manager:

    >>> with reapy_boost.undo_block('add track'):
    ...     reapy_boost.Project.add_track()

    >>> @reapy_boost.undo_block('add track')
    >>> def some_function(*args, **kwargs):
    ...     reapy_boost.Project.add_track()

    :param undo_name: Str to register undo name (shown later in Undo menu)
    :param flags: Int to pass to Undo_EndBlock
                    (leave default if you don't know what it is)
        1: track configurations
        2: track FX
        4: track items
        8: project states
        16: freeze states
    """

    def __init__(self, undo_name: str, flags: int = -1) -> None:
        self.undo_name = undo_name
        self.flags = flags

    def __enter__(self) -> None:
        RPR.Undo_BeginBlock()  # type:ignore

    def __exit__(
        self, exc_type: Exception, exc_val: str, exc_tb: TracebackType
    ) -> None:
        RPR.Undo_EndBlock(self.undo_name, self.flags)  # type:ignore


def update_arrange() -> None:
    """
    Redraw the arrange view.
    """
    RPR.UpdateArrange()  # type:ignore


def update_timeline() -> None:
    """
    Redraw the arrange view and ruler.
    """
    RPR.UpdateTimeline()  # type:ignore


def view_prefs() -> None:
    """
    Open Preferences.
    """
    RPR.ViewPrefs(0, "")  # type:ignore



from typing import Dict, Optional, Tuple
import reapy_boost
from reapy_boost import reascript_api as RPR
from reapy_boost.core import ReapyObject


class TimeSelection(ReapyObject):

    _class_name = "TimeSelection"

    def __init__(
        self,
        parent_project: Optional['reapy_boost.Project'] = None,
        parent_project_id: Optional[str] = None
    ):
        if parent_project_id is None:
            message = (
                "One of `parent_project` or `parent_project_id` must be "
                "specified."
            )
            assert parent_project is not None, message
            parent_project_id = parent_project.id
        self.project_id: str = parent_project_id

    def _get_infos(self) -> Tuple[str, bool, bool, float, float, bool]:
        """
        Return infos as returned by RPR.GetSet_LoopTimeRange2.

        Returns
        -------
        infos : tuple
            str: proj_id
            bool: isSet (False)
            bool: isLoop
            float: start
            float: end
            bool: allowautoseek
        """
        infos = RPR.GetSet_LoopTimeRange2(  # type:ignore
            self.project_id, False, False, 0, 0, False
        )
        return infos  # type:ignore

    def _set_start_end(
        self,
        start: Optional[float] = None,
        end: Optional[float] = None
    ) -> None:
        infos = list(
            RPR.GetSet_LoopTimeRange2(  # type:ignore
                self.project_id, False, False, 0, 0, False
            )
        )
        if start is None:
            start = infos[3]
        if end is None:
            end = infos[4]
        infos[1], infos[3], infos[4] = True, start, end
        RPR.GetSet_LoopTimeRange2(*infos)  # type:ignore

    @property
    def _kwargs(self) -> Dict[str, str]:  # type:ignore
        return {"parent_project_id": self.project_id}

    @property
    def end(self) -> float:
        """
        Return time selection end in seconds.

        Returns
        -------
        end : float
            Time selection end in seconds.
        """
        return self._end_inside()

    @reapy_boost.inside_reaper()
    def _end_inside(self) -> float:
        infos = self._get_infos()
        end = infos[4]
        return end

    @end.setter
    def end(self, end: float) -> None:
        """
        Set time selection end.

        Parameters
        ----------
        end : float
            Time selection end in seconds.
        """
        self._set_start_end(end=end)

    @property
    def is_looping(self) -> bool:
        """
        Return whether looping is enabled.

        Returns
        -------
        looping : bool
            Whether looping is enabled.
        """
        is_looping = bool(
            RPR.GetSetRepeatEx(  # type:ignore
                self.project_id, -1
            )
        )
        return is_looping

    @is_looping.setter
    def is_looping(self, is_looping: bool) -> None:
        """
        Sets whether time selection should loop.

        Parameters
        ----------
        looping : bool
            Whether time selection should loop.
        """
        if is_looping:
            self.loop()
        else:
            self.unloop()

    @property
    def length(self) -> float:
        """
        Return time selection length in seconds.

        Returns
        -------
        length : float
            Time selection length in seconds.
        """
        return self._length_inside()

    @reapy_boost.inside_reaper()
    def _length_inside(self) -> float:
        infos = self._get_infos()
        start, end = infos[3:5]
        length = end - start
        return length

    @length.setter
    def length(self, length: float) -> None:
        """
        Set time selection length (by moving its end).

        Parameters
        ----------
        length : float
            Time selection length in seconds.
        """
        infos = list(
            RPR.GetSet_LoopTimeRange2(  # type:ignore
                self.project_id, False, False, 0, 0, False
            )
        )
        infos[1], infos[4] = True, infos[3] + length
        RPR.GetSet_LoopTimeRange2(*infos)  # type:ignore

    def loop(self) -> None:
        """
        Enable time selection looping.

        See also
        --------
        TimeSelection.is_looping
        TimeSelection.unloop
        """
        RPR.GetSetRepeatEx(self.project_id, 1)  # type:ignore

    @property
    def start(self) -> float:
        """
        Return time selection start in seconds.

        Returns
        -------
        start : float
            Time selection start in seconds.
        """
        return self._start_inside()

    @reapy_boost.inside_reaper()
    def _start_inside(self) -> float:
        infos = self._get_infos()
        start = infos[3]
        return start

    @start.setter
    def start(self, start: float) -> None:
        """
        Set time selection start.

        Parameters
        ----------
        start : float
            New time selection start.
        """
        self._set_start_end(start)

    def shift(self, direction: str = "") -> None:
        """
        Shift time selection.

        Parameters
        ----------
        direction : {"right", "left"}
            Direction to which time selection will be shifted. Nothing
            happens if direction is neither "right" nor "left". Note
            that the shift size depends on whether snap is enabled
            and of the zoom level.
        """
        if direction == "right":
            RPR.Loop_OnArrow(self.project_id, 1)  # type:ignore
        elif direction == "left":
            RPR.Loop_OnArrow(self.project_id, -1)  # type:ignore

    def unloop(self) -> None:
        """
        Disable time selection looping.

        See also
        --------
        TimeSelection.is_looping
        TimeSelection.loop
        """
        RPR.GetSetRepeatEx(self.project_id, 0)  # type:ignore
